/*
	SourceGear Vault
	Copyright 2002-2003 SourceGear Corporation
	All Rights Reserved.
	
	You may not distribute this code, or any portion thereof, 
	or any derived work thereof, neither in source code form 
	nor in compiled form, to anyone outside your organization.
	
	The opslib API is currently ready only for the somewhat
	adventuresome.  The following disclaimers apply:
	
	1.  The API is not frozen.  If you write code to the opslib
		API, you may need to update your code the next time we
		release.
		
	2.  The API is not well documented.  (Corollary:  This
		sample code is not well commented.)
	
	3.  The only support we can offer is currently on our
		public discussion mailing lists.  Information about
		how to subscribe to our mailing lists is available
		in the Vault section of our website:
		
			http://www.sourcegear.com/vault/index.asp
	
	We are continually working to improve the usability of this
	API for our customers.  Your feedback is always welcome.
	
*/

using System;
using System.IO;
using System.Collections;
using System.Text;
using System.Text.RegularExpressions;
using System.Xml;

using VaultClientOperationsLib;
using VaultClientNetLib;
using VaultClientNetLib.ClientService;
using VaultLib;

namespace VaultCmdLineClient
{
	class VaultCmdLineClient
	{
		#region MemberVariables
		private const string SESSION_FILENAME = "vault_cmdline_client_session.txt";

		private ClientInstance _ci = null;
		private Args _args = null;
		private XMLOutputWriter _xml = null;
		#endregion

		#region PublicMethods
		
		[STAThread]
		public static int Main(string[] args)
		{
			int nRetCode = 0;
			bool bOK = false;

			// create the args.
			Args cmdlineargs = new Args(args);

			// create the output writer - based on the OUT option
			XMLOutputWriter xml = new XMLOutputWriter(cmdlineargs.Out);

			// create the cmd line client object
			VaultCmdLineClient cmdlineclient = new VaultCmdLineClient(cmdlineargs, xml);

			// pre process check on the original command
			if ( cmdlineargs.Error == false )
			{
				cmdlineclient.PreProcessCommand(cmdlineargs);
			}

			cmdlineargs.Out.WriteLine("<vault>");
			if(cmdlineargs.Error)
			{
				xml.Begin("error");
				xml.WriteContent(cmdlineargs.ErrorMessage);
				xml.End();
			}
			else
			{
				try
				{
					// if there is a batch, read each line until no more commands
					if ( cmdlineargs.InBatchMode == false )
					{
						// process the one command.
						bOK = cmdlineclient.ProcessCommand(cmdlineargs);
					}
					else
					{
						// while there are commands on the batch input stream
						// keep processing.
						string strCmd = null;
						while ( null != (strCmd = cmdlineargs.BatchTextReader.ReadLine()) )
						{
							if ( strCmd.Length > 0 )
							{
								Args batch_args = cmdlineargs.CreateBatchCmdArgs();

								// parse / and process the batch command
								batch_args.ParseBatchCommand(strCmd);
								bOK = cmdlineclient.ProcessCommand(batch_args);
								if ( bOK == false )
								{
									break;
								}
							}
						}
					}
				}
				catch (UsageException e)
				{
					xml.Begin("error");
					xml.WriteContent(e.Message);
					xml.End();
				}
				catch (Exception e)
				{
					xml.Begin("error");
					xml.WriteContent(e.Message);
					xml.End();
					xml.Begin("exception");
					xml.WriteContent(e.ToString());
					xml.End();
				}
				finally
				{
					// always force a logout here.
					cmdlineclient.Logout(true);
				}
			}

			xml.Begin("result");
			xml.AddPair("success", bOK);
			xml.End();

			cmdlineargs.Out.WriteLine("</vault>");

			// clean up any left over streams
			cmdlineargs.CloseOutputStream();
			cmdlineargs.CloseInputBatchStream();

			if ( bOK == false )
			{
				nRetCode = -1;
			}

			return nRetCode;
		}

		internal static bool CommandNeedsLogin(Command c)
		{
			bool bRet = true;
			switch(c)
			{
				case Command.HELP:
				case Command.NONE:
				case Command.INVALID:
				case Command.FORGETLOGIN:
					bRet = false;
					break;
			}
			return bRet;
		}
		
		internal static bool CommandNeedsRepositorySpecified(Command c)
		{
			bool bRet = true;
			switch (c)
			{
				case Command.ADDREPOSITORY:
				case Command.ADDUSER:
				case Command.LISTUSERS:
				case Command.FORGETLOGIN:
				case Command.HELP:
				case Command.NONE:
				case Command.INVALID:
				case Command.LISTREPOSITORIES:
					bRet = false;
					break;
			}
			return bRet;
		}

		internal static Option LookupOptionByString(string strOption)
		{
			Option oRet = Option.INVALID;

			foreach (Option o in Enum.GetValues(typeof(Option)))
			{
				// Note that our cmdline options are case-INsensitive.
				// Most UNIX apps have cmdline options which are case-sensitive
				bool bCaseInsensitive = true;
				if (string.Compare(strOption, o.ToString(), bCaseInsensitive) == 0)
				{
					oRet = o;
					break;
				}
			}
			return oRet;
		}
		
		internal static bool CommandNeedsAdmin(Command c)
		{
			bool bRet = false;
			switch (c)
			{
				case Command.ADDREPOSITORY:
				case Command.ADDUSER:
				case Command.LISTUSERS:
					bRet = true;
					break;
			}
			return bRet;
		}
		internal static MergeOption LookupMergeOptionByString(string strMergeOption)
		{
			MergeOption moRet = MergeOption.none;

			foreach (MergeOption mo in Enum.GetValues(typeof(MergeOption)))
			{
				// Note that our cmdline options are case-INsensitive.
				// Most UNIX apps have cmdline options which are case-sensitive
				bool bCaseInsensitive = true;
				if (string.Compare(strMergeOption, mo.ToString(), bCaseInsensitive) == 0)
				{
					moRet = mo;
					break;
				}
			}
			return moRet;
		}
		
		internal static FileTimeOption LookupFileTimeOptionByString(string strFileTimeOption)
		{
			FileTimeOption ftoRet = FileTimeOption.none;

			foreach (FileTimeOption fto in Enum.GetValues(typeof(FileTimeOption)))
			{
				// Note that our cmdline options are case-INsensitive.
				// Most UNIX apps have cmdline options which are case-sensitive
				bool bCaseInsensitive = true;
				if (string.Compare(strFileTimeOption, fto.ToString(), bCaseInsensitive) == 0)
				{
					ftoRet = fto;
					break;
				}
			}
			return ftoRet;
		}
		#endregion

		#region PrivateMethods
		private void Login(bool bAllowAuto, bool bSaveSession)
		{
			bool bResult = false;

			// get access level req'd by the command
			VaultConnection.AccessLevelType altCommand = (CommandNeedsAdmin(_args.Cmd) == false) ? VaultConnection.AccessLevelType.Client : VaultConnection.AccessLevelType.Admin;

			if ( _ci != null ) 
			{
				if ( (_ci.ConnectionStateType == ConnectionStateType.Connected) && 
					(_ci.AccessLevel == altCommand) )
				{
					// already logged in
					return;
				}
				else
				{
					_ci.Logout();
				}
			}

			// init the client instance
			_ci = new ClientInstance();
			_ci.UseFileSystemWatchers = false;
			_ci.Init(altCommand);

			// get info for logging in
			string strURL = _args.Url;
			string strUser = _args.User;
			string strPassword = _args.Password;
			string strRepository = _args.Repository;
			if (
				(bAllowAuto == true) && 
				((strURL == null) || (strUser == null))
				)
			{
				RetrieveSession(ref strURL, ref strUser, ref strPassword, ref strRepository);
			}

			// lower case - used later in the method
			if ( strRepository != null )
			{
				strRepository = strRepository.ToLower();
			}

			if (
				(strURL != null) && 
				(strUser != null)
				)
			{
				if (strRepository == null)
				{
					if ( CommandNeedsRepositorySpecified(_args.Cmd) == true )
					{
						throw new UsageException(string.Format("You must specify a repository for the {0} command", _args.Cmd));
					}
				}
				
				try
				{
					_ci.Login(strURL, strUser, strPassword);
				}
				catch(Exception e)
				{
					throw new Exception( string.Format("The connection to the server failed: server cannot be contacted or uses a protocol that is not supported by this client. {0}", e.Message) ); 
				}

				if (_ci.ConnectionStateType != ConnectionStateType.Connected)
				{
					throw new Exception("Login failed.");
				}
				else
				{
					bResult = true;

					if (strRepository != null)
					{
						VaultRepositoryInfo theRep = null;

						VaultRepositoryInfo[] reps = null;
						_ci.ListRepositories(ref reps);

						foreach (VaultRepositoryInfo r in reps)
						{
							if (r.RepName.ToLower() == strRepository)
							{
								theRep = r;
								break;
							}
						}

						if (theRep != null)
						{
							_ci.SetActiveRepositoryID(theRep.RepID, strUser, theRep.UniqueRepID, true, true);
						}
						else
						{
							throw new UsageException(string.Format("Repository {0} not found", strRepository));
						}
					}

					// set working folder options
					_ci.WorkingFolderOptions.RequireCheckOutBeforeCheckIn = _args.RequireCheckOut;
					_ci.WorkingFolderOptions.DefaultLocalCopyType = _args.LocalCopy;

					if ( (bAllowAuto == false ) &&
						(bSaveSession) )
					{
						StoreSession(strURL, strUser, strPassword, strRepository);
					}
				}
			}
			else
			{
				throw new UsageException("Please specify -user, -password, -host, and -repository.");
			}

			if ( _ci != null )
			{
				_ci.NewMessage += new NewMessageEventHandler(NewMessageHandler);
				_ci.NewBulkMessages += new NewBulkMessagesEventHandler(NewBulkMessagesHandler);
			}

			if (bResult == false)
			{
				throw new UsageException("Please verify that you have specified -user, -password, -host, and -repository.");
			}
		}
		
		private void Login()
		{
			Login(true, false);
		}

		private void Logout(bool bForceLogout)
		{
			if ( (_ci != null) && 
				((bForceLogout == true) || (_args.InBatchMode == false))
				)
			{
				// log out when forced or not in batch mode.
				_ci.Logout();
				_ci = null;
			}
		}

		private bool ParseBool(string s)
		{
			bool bRet = false;

			switch (s.ToLower().Trim())
			{
				case "yes":
				case "true":
					bRet = true;
					break;
			}
			return bRet;
		}
		
		private static Command LookupCommandByString(string strCmd)
		{
			Command cmdRet = Command.NONE;

			foreach (Command c in Enum.GetValues(typeof(Command)))
			{
				// Note that our cmdline options are case-INsensitive.
				// Most UNIX apps have cmdline options which are case-sensitive
				bool bCaseInsensitive = true;
				if (string.Compare(strCmd, c.ToString(), bCaseInsensitive) == 0)
				{
					cmdRet = c;
					break;
				}
			}
			return cmdRet;
		}
		
		private static CompareToOption LookupCompareToOptionByString(string strCompareToOption)
		{
			CompareToOption ctoRet = CompareToOption.current;

			foreach (CompareToOption cto in Enum.GetValues(typeof(CompareToOption)))
			{
				// Note that our cmdline options are case-INsensitive.
				// Most UNIX apps have cmdline options which are case-sensitive
				bool bCaseInsensitive = true;
				if (string.Compare(strCompareToOption, cto.ToString(), bCaseInsensitive) == 0)
				{
					ctoRet = cto;
					break;
				}
			}
			return ctoRet;
		}

		private string GetStatusString(WorkingFolderFileStatus st)
		{
			return (st != WorkingFolderFileStatus.None) ? st.ToString() : string.Empty;
		}

		private void NewMessageHandler(object sender, ProgressMessage message)
		{
			_xml.WriteUserMessage(message.Message);
		}

		private void NewBulkMessagesHandler(object sender, ArrayList aProgressMessages)
		{
			if (_args.Verbose)
			{
				foreach (ProgressMessage message in aProgressMessages)
				{
					_xml.WriteUserMessage(message.Message);
				}
			}
		}

		private void WriteListWorkingFolders()
		{
			string[] repos = null;
			string[] disk = null;

			_ci.TreeCache.GetWorkingFolderAssignments(ref repos, ref disk);

			_xml.Begin("listworkingfolders");
			for (int i=0; i<repos.Length; i++)
			{
				_xml.Begin("workingfolder");
				_xml.AddPair("reposfolder", repos[i]);
				_xml.AddPair("localfolder", disk[i]);
				_xml.End();
			}
			_xml.End();
		}

		private void WriteFolder(VaultClientFolder vcfolder, bool recursive, int depth)
		{
			WorkingFolder wf = _ci.GetWorkingFolder(vcfolder);

			_xml.Begin("folder");
			_xml.AddPair("name", vcfolder.FullPath);
			if (wf != null)
			{
				_xml.AddPair("workingfolder", wf.GetLocalFolderPath());
			}
			foreach (VaultClientFile file in vcfolder.Files)
			{
				_xml.Begin("file");
				_xml.AddPair("name", file.Name);
				_xml.AddPair("version", file.Version);
				_xml.AddPair("length", file.FileLength);
				_xml.AddPair("objectid", file.ID);
				_xml.AddPair("objectversionid", file.ObjVerID);

				string strCheckOuts = _ci.GetCheckOuts(file);
				if (
					(strCheckOuts != null)
					&& (strCheckOuts.Length > 0)
					)
				{
					_xml.AddPair("checkouts", strCheckOuts);
				}

				if (wf != null)
				{
					WorkingFolderFileStatus st = wf.GetStatus(file);
					if (st != WorkingFolderFileStatus.None)
					{
						_xml.AddPair("status", GetStatusString(st));
					}
				}

				_xml.End();
			}

			foreach (VaultClientFolder subfolder in vcfolder.Folders)
			{
				if (recursive)
				{
					WriteFolder(subfolder, recursive, depth+1);
				}
			}
			_xml.End();
		}

		private void WriteChangeSet()
		{
			_ci.UpdateKnownChanges_All(false);
			ChangeSetItemColl csic = _ci.InternalChangeSet_GetItems(true);
			WriteChangeSet(csic);
		}

		private void WriteChangeSet(ChangeSetItemColl csic)
		{
			if ( _xml != null )
			{
				_xml.Begin("changeset");
			}

			try
			{
				if ( (csic != null) && (csic.Count > 0) )
				{
					int i, nCnt;
					ChangeSetItem csi = null;
					for (i = 0, nCnt = csic.Count; i<nCnt; i++)
					{
						csi = csic[i];
						switch (csi.Type)
						{
								#region ChangeSets
							case ChangeSetItemType.AddFile:
							{
								ChangeSetItem_AddFile it = (ChangeSetItem_AddFile) csi;
							
								_xml.Begin("AddFile");
								_xml.AddPair("id", i.ToString());
								_xml.AddPair("repospath", it.DisplayRepositoryPath);
								_xml.AddPair("localpath", it.DiskFile);
								_xml.End();
							
								break;
							}
							case ChangeSetItemType.AddFolder:
							{
								ChangeSetItem_AddFolder it = (ChangeSetItem_AddFolder) csi;

								_xml.Begin("AddFolder");
								_xml.AddPair("id", i.ToString());
								_xml.AddPair("reposfolder", it.DisplayRepositoryPath);
								_xml.AddPair("localfolder", it.DiskFolder);
								_xml.End();

								break;
							}
							case ChangeSetItemType.DeleteFile:
							{
								ChangeSetItem_DeleteFile it = (ChangeSetItem_DeleteFile) csi;

								_xml.Begin("DeleteFile");
								_xml.AddPair("id", i.ToString());
								_xml.AddPair("repospath", it.DisplayRepositoryPath);
								_xml.End();

								break;
							}
							case ChangeSetItemType.DeleteFolder:
							{
								ChangeSetItem_DeleteFolder it = (ChangeSetItem_DeleteFolder) csi;

								_xml.Begin("DeleteFolder");
								_xml.AddPair("id", i.ToString());
								_xml.AddPair("repospath", it.DisplayRepositoryPath);
								_xml.End();

								break;
							}
							case ChangeSetItemType.CreateFolder:
							{
								ChangeSetItem_CreateFolder it = (ChangeSetItem_CreateFolder) csi;

								_xml.Begin("CreateFolder");
								_xml.AddPair("id", i.ToString());
								_xml.AddPair("repospath", it.DisplayRepositoryPath);
								_xml.End();

								break;
							}
							case ChangeSetItemType.Branch:
							{
								ChangeSetItem_CopyBranch it = (ChangeSetItem_CopyBranch) csi;

								_xml.Begin("Branch");
								_xml.AddPair("id", i.ToString());
								_xml.AddPair("repospath", it.DisplayRepositoryPath);
								_xml.AddPair("branchpath", it.BranchPath);
								_xml.End();
							
								break;
							}
							case ChangeSetItemType.Share:
							{
								ChangeSetItem_Share it = (ChangeSetItem_Share) csi;

								_xml.Begin("Share");
								_xml.AddPair("id", i.ToString());
								_xml.AddPair("repospath", it.RepositoryPath);
								_xml.AddPair("sharepath", it.NewSharePath);
								_xml.End();

								break;
							}
							case ChangeSetItemType.Pin:
							{
								ChangeSetItem_Pin it = (ChangeSetItem_Pin) csi;

								_xml.Begin("Pin");
								_xml.AddPair("id", i.ToString());
								_xml.AddPair("repospath", it.DisplayRepositoryPath);
								_xml.End();

								break;
							}
							case ChangeSetItemType.Rename:
							{
								ChangeSetItem_Rename it = (ChangeSetItem_Rename) csi;

								_xml.Begin("Rename");
								_xml.AddPair("id", i.ToString());
								_xml.AddPair("repospath", it.OldRepositoryPath);
								_xml.AddPair("newname", it.NewName);
								_xml.End();
								break;
							}
							case ChangeSetItemType.Unpin:
							{
								ChangeSetItem_Unpin it = (ChangeSetItem_Unpin) csi;

								_xml.Begin("Unpin");
								_xml.AddPair("id", i.ToString());
								_xml.AddPair("repospath", it.DisplayRepositoryPath);
								_xml.End();

								break;
							}
							case ChangeSetItemType.Move:
							{
								ChangeSetItem_Move it = (ChangeSetItem_Move) csi;

								_xml.Begin("Move");
								_xml.AddPair("id", i.ToString());
								_xml.AddPair("repospath", it.OldRepositoryPath);
								_xml.AddPair("newpath", it.NewOwnerPath);
								_xml.End();

								break;
							}
							case ChangeSetItemType.Modified:
							{
								ChangeSetItem_Modified it = (ChangeSetItem_Modified) csi;

								_xml.Begin("ModifyFile");
								_xml.AddPair("id", i.ToString());
								_xml.AddPair("repospath", it.DisplayRepositoryPath);
								_xml.AddPair("localpath", it.DiskFile);
								_xml.End();

								break;
							}
							case ChangeSetItemType.Unmodified:
							{
								ChangeSetItem_Unmodified it = (ChangeSetItem_Unmodified) csi;

								_xml.Begin("UnmodifiedFile");
								_xml.AddPair("id", i.ToString());
								_xml.AddPair("respospath", it.DisplayRepositoryPath);
								_xml.AddPair("localpath", it.DiskFile);
								_xml.End();

								break;
							}
							case ChangeSetItemType.ChangeExtProperties:
							{
								ChangeSetItem_ChangeExtProperties it = (ChangeSetItem_ChangeExtProperties) csi;

								_xml.Begin("ChangeExtProperties");
								_xml.AddPair("id", i.ToString());
								_xml.AddPair("repospath", it.DisplayRepositoryPath);
								_xml.End();

								break;
							}
							case ChangeSetItemType.CheckedOutMissing:
							{
								ChangeSetItem_CheckedOutMissing it = (ChangeSetItem_CheckedOutMissing) csi;

								_xml.Begin("CheckedOutMissing");
								_xml.AddPair("id", i.ToString());
								_xml.AddPair("respospath", it.DisplayRepositoryPath);
								_xml.End();

								break;
							}
							case ChangeSetItemType.Snapshot:
							{
								ChangeSetItem_Snapshot it = (ChangeSetItem_Snapshot)csi;
								_xml.Begin("Snapshot");
								_xml.AddPair("id", i.ToString());
								_xml.AddPair("respospath", it.RepositoryPath);
								_xml.AddPair("parentpath", it.SnapshotPath);
								_xml.End();

								break;
							}
							default:
							{
								// this should never happen.
								throw new Exception("ACK!  There is a ChangeSet item we don't recognize.");
							}
								#endregion
						}
					}
				}
			}
			finally
			{
				// end the xml pair.
				if ( _xml != null )
				{
					_xml.End();
				}
			}

		}

		private void WriteCheckOuts(VaultClientCheckOutItem[] checkOuts)
		{
			_xml.Begin("checkoutlist");

			if ( (checkOuts != null) && (checkOuts.Length > 0) )
			{
				foreach (VaultClientCheckOutItem item in checkOuts)
				{
					_xml.Begin("checkoutitem");
					_xml.AddPair("id", item.FileID);
					
					foreach (VaultClientCheckOutUser user in item.CheckOutUsers)
					{
						_xml.Begin("checkoutuser");
						_xml.AddPair("username", user.Name);
						_xml.AddPair("version", user.Version);
						_xml.AddPair("repositorypath", user.RepPath);

						switch (user.LockType)
						{
							case VaultCheckOutType.None:
								_xml.AddPair("locktype", "none");
								break;
							case VaultCheckOutType.CheckOut:
								_xml.AddPair("locktype", "checkout");
								break;
							case VaultCheckOutType.Exclusive:
								_xml.AddPair("locktype", "exclusive");
								break;
							default:
								_xml.AddPair("locktype", "unknown");
								break;
						}

						_xml.AddPair("comment", user.Comment);
						_xml.AddPair("hostname", user.Hostname);
						_xml.AddPair("localpath", user.LocalPath);
						_xml.AddPair("folderid", user.FolderID);
						_xml.AddPair("lockedwhen", user.LockedWhen.ToString());
						_xml.AddPair("miscinfo", user.MiscInfo);
						_xml.End();
					}
					
					_xml.End();
				}
			}

			_xml.End();
		}

		private string GetHistItemTypeString(int x)
		{
			return x.ToString();
		}

		private void ValidateReposPath(string s)
		{
			if (s[0] != '$')
			{
				throw new UsageException(string.Format("Invalid repository path: {0}", s));
			}

			if (s.Length > 1)
			{
				if (s[1] != '/')
				{
					throw new UsageException(string.Format("Invalid repository path: {0}", s));
				}
			}
		}

		private bool ValidateChangeSetItemID(int nID)
		{
			ChangeSetItemColl csic = _ci.InternalChangeSet_GetItems(true);
			return ValidateChangeSetItemID(nID, csic);
		}

		private bool ValidateChangeSetItemID(int nID, ChangeSetItemColl csic)
		{
			// get the change set.
			bool bValid = false;
			if ( csic != null )
			{
				bValid = ( (nID >= 0) && (nID < csic.Count) );
			}
			return bValid;
		}

		private void DoListUsers()
		{
			VaultClientNetLib.AdminService.VaultUser[] users = null;
			_ci.Connection.ListUsers(ref users);

			// look up ID for repository specified
			int reposID = GetRepositoryId( _args.Repository );

			// sort users before outputting to XML stream
			Array.Sort(users, new UserItemComparer());

			_xml.Begin("listusers");
			foreach (VaultClientNetLib.AdminService.VaultUser u in users)
			{
				_xml.Begin("user");
				_xml.AddPair("login", u.Login);
				if ( _args.Verbose )
				{
					_xml.AddPair("email", u.Email);
					_xml.AddPair("active", u.isActive);

					// groups
					_xml.Begin("groups");
					VaultClientNetLib.AdminService.VaultGroup[] groups = u.BelongToGroups;

					// sort groups before outputting to XML stream
					Array.Sort(groups, new GroupComparer() );

					foreach ( VaultClientNetLib.AdminService.VaultGroup aGroup in groups )
					{
						_xml.Begin("group");
						_xml.WriteContent(aGroup.Name);
						_xml.End();
					}
					_xml.End();

					// rights
					_xml.AddPair("defaultRights", DecodeUserRights(u.DefaultRights) );
					_xml.Begin("folderRights");
					VaultClientNetLib.AdminService.VaultFolderRightsItem[] rights = null;
					_ci.Connection.ListRightsByUser(u.UserID, ref rights);

					// sort rights before outputting XML stream
					Array.Sort(rights, new RightComparer() );

					if (rights.Length > 0)
					{
						foreach ( VaultClientNetLib.AdminService.VaultFolderRightsItem anItem in rights )
						{
							if ( anItem.RepID == reposID )
							{
								_xml.Begin("singleRight");
								_xml.AddPair("folder", anItem.Path);
								_xml.AddPair("right", DecodeUserRights(anItem.FolderRights));
								_xml.End();
							}
						}
					}
					_xml.End();
				}
				_xml.End();
			}
			_xml.End();
		}

		private void x_emitOpItem(string tag, VaultClientFile f)
		{
			_xml.Begin(tag);
			_xml.AddPair("fullpath", f.FullPath);
			_xml.AddPair("version", f.Version);
			_xml.End();
		}

		private void x_emitOpItem(string tag, VaultClientFolder f)
		{
			_xml.Begin(tag);
			_xml.AddPair("fullpath", f.FullPath);
			_xml.AddPair("version", f.Version);
			_xml.End();
		}

		private bool ProcessCommand(Args curArg)
		{
			bool bSuccess = true;

			// assign the new set of arguments
			_args = curArg;

			try
			{
				#region commands
				switch (curArg.Cmd)
				{
					case Command.ADD:
					{
						// The first item must be the repository folder
						// All other items are local paths to be added
						if (curArg.items.Count < 2)
						{
							throw new UsageException("usage: ADD repository_folder path_to_add [...]");
						}

						string strReposFolder = (string) curArg.items[0];

						ArrayList strItemArray = new ArrayList();
						for (int i=1; i < curArg.items.Count; i++)
						{
							strItemArray.Add(curArg.items[i]);
						}

						bSuccess = ProcessCommandAdd(strReposFolder, strItemArray);
						break;
					}

					case Command.ADDREPOSITORY:
					{
						if (curArg.items.Count != 1)
						{
							throw new UsageException("usage: ADDREPOSITORY repository_folder");
						}

						string strNewReposName = (string) curArg.items[0];

						bSuccess = ProcessCommandAddRepository(strNewReposName);
						break;
					}

					case Command.ADDUSER:
					{
						if ( (curArg.items.Count == 0) || (curArg.items.Count > 3) )
						{
							throw new UsageException("usage: ADDUSER login [password] [email]");
						}

						string strLogin = (string) curArg.items[0];

						string strPassword = string.Empty;
						if ( curArg.items.Count > 1 )
						{
							strPassword = (string) curArg.items[1];
						}

						string strEmail = null;
						if ( curArg.items.Count == 3 )
						{
							strEmail = (string) curArg.items[2];
						}

						bSuccess = ProcessCommandAddUser(strLogin, strPassword, strEmail);
						break;
					}
					case Command.BATCH:
					{
						throw new Exception( string.Format("{0} cannot be called as a batch command", curArg.Cmd) );
					}
					case Command.BLAME:
					{
						if (curArg.items.Count != 2)
						{
							throw new UsageException("usage: BLAME path linenumber");
						}
						string strReposPath = (string) curArg.items[0];
						ValidateReposPath(strReposPath);

						int linenumber = int.Parse((string)curArg.items[1]);

						bSuccess = ProcessCommandBlame(strReposPath, linenumber);
						break;
					}

					case Command.BRANCH:
					{
						if (curArg.items.Count != 2)
						{
							throw new UsageException("usage: BRANCH from_path to_path");
						}

						string strReposPath_From = (string) curArg.items[0];
						ValidateReposPath(strReposPath_From);

						string strReposPath_To = (string) curArg.items[1];
						ValidateReposPath(strReposPath_To);

						bSuccess = ProcessCommandBranch(strReposPath_From, strReposPath_To);
						break;
					}

					case Command.CHECKOUT:
					{
						if (curArg.items.Count < 1)
						{
							throw new UsageException("usage: CHECKOUT item [...]");
						}

						ArrayList strItemArray = new ArrayList();
						foreach (string strReposItem in curArg.items)
						{
							ValidateReposPath(strReposItem);
							strItemArray.Add(strReposItem);
						}

						bSuccess = ProcessCommandCheckout(strItemArray);
						break;
					}
					case Command.CLOAK:
					{
						if (curArg.items.Count < 1)
						{
							throw new UsageException("usage: CLOAK item [...]");
						}
						
						ArrayList strItemArray = new ArrayList();
						foreach (string strReposItem in curArg.items)
						{
							ValidateReposPath(strReposItem);
							strItemArray.Add(strReposItem);
						}

						bSuccess = ProcessCommandCloak(strItemArray);
						break;
					}

					case Command.COMMIT:
					{
						ArrayList strItemArray = new ArrayList();
						foreach (string strReposItem in curArg.items)
						{
							strItemArray.Add(strReposItem);
						}

						bSuccess = ProcessCommandCommit(strItemArray);
						break;
					}

					case Command.CREATEFOLDER:
					{
						if (curArg.items.Count != 1)
						{
							throw new UsageException("usage: CREATEFOLDER foldername");
						}

						string strReposFolder = (string) curArg.items[0];
						ValidateReposPath(strReposFolder);

						bSuccess = ProcessCommandCreateFolder(strReposFolder);
						break;
					}

					case Command.DELETE:
					{
						if (curArg.items.Count < 1)
						{
							throw new UsageException("usage: DELETE item [...]");
						}
						
						ArrayList strItemArray = new ArrayList();
						foreach (string strReposItem in curArg.items)
						{
							ValidateReposPath(strReposItem);
							strItemArray.Add(strReposItem);
						}

						bSuccess = ProcessCommandDelete(strItemArray);
						break;
					}

					case Command.DIFF:
					{
						string strRepItem = null, strRepLbl = null, 
							strDiffAgainstItem = null, strDiffAgainstLbl = null, 
							strErrMsg = null;

						// the "compare to" operator
						DiffAgainstType datDiffChoice = DiffAgainstType.CurrentRepositoryVersion;
						CompareToOption cto = VaultCmdLineClient.LookupCompareToOptionByString(curArg.DiffCompareTo);
						switch (cto)
						{
							case CompareToOption.current:

								if ( curArg.items.Count == 1 )
								{
									datDiffChoice = DiffAgainstType.CurrentRepositoryVersion;
									strRepItem = (string)curArg.items[0];
									// TODO - move into resource for globalization
									strRepLbl = "Working: {0}";
									strDiffAgainstLbl = "Repository: {0}";
								}
								else
								{
									strErrMsg = "usage: DIFF repository_path [...]";
								}
								break;

							case CompareToOption.label:

								if ( curArg.items.Count == 2 )
								{
									datDiffChoice = DiffAgainstType.Label;
									strRepItem = (string)curArg.items[0];
									strDiffAgainstItem = (string)curArg.items[1];

									// TODO - move into resource for globalization
									strRepLbl = "Label: {0}";
									strDiffAgainstLbl = "Working: {0}";
								}
								else
								{
									strErrMsg = "usage: DIFF repository_path label_name [...]";
								}
								break;

							case CompareToOption.lastget:

								if ( curArg.items.Count == 1 )
								{
									datDiffChoice = DiffAgainstType.PreviousRepositoryVersion;
									strRepItem = (string)curArg.items[0];

									// TODO - move into resource for globalization
									strRepLbl = "Baseline: {0}";
									strDiffAgainstLbl = "Working Version: {0}";
								}
								else
								{
									strErrMsg = "usage: DIFF repository_path [...]";
								}
								break;

							case CompareToOption.local:

								if ( curArg.items.Count == 2 )
								{
									datDiffChoice = DiffAgainstType.AnyLocalItem;
									strRepItem = (string)curArg.items[0];
									strDiffAgainstItem = (string)curArg.items[1];

									// TODO - move into resource for globalization
									strRepLbl = "{0}";
									strDiffAgainstLbl = "Working: {0}";
								}
								else
								{
									strErrMsg = "usage: DIFF repository_path local_path [...]";
								}
								break;

							case CompareToOption.repository:

								if ( curArg.items.Count == 2 )
								{
									datDiffChoice = DiffAgainstType.AnyRepositoryItem;
									strRepItem = (string)curArg.items[0];
									strDiffAgainstItem = (string)curArg.items[1];

									// TODO - move into resource for globalization
									strRepLbl = "Repository: {0}";
									strDiffAgainstLbl = "Working: {0}";
								}
								else
								{
									strErrMsg = "usage: DIFF repository_path other_repository_path [...]";
								}
								break;

							default:

								if ( curArg.items.Count == 1 )
								{
									datDiffChoice = DiffAgainstType.CurrentRepositoryVersion;
									strRepItem = (string)curArg.items[0];

									// TODO - move into resource for globalization
									strRepLbl = "Working: {0}";
									strDiffAgainstLbl = "Repository: {0}";
								}
								else
								{
									strErrMsg = "usage: DIFF repository_path [...]";
								}
								break;
						}

						if ( strErrMsg != null )
						{
							throw new UsageException(strErrMsg);
						}

						bSuccess = ProcessCommandDiff(curArg.DiffBin, curArg.DiffArgs, datDiffChoice, 
							curArg.Recursive, strRepItem, strRepLbl, strDiffAgainstItem, strDiffAgainstLbl);
						break;
					}

					case Command.FORGETLOGIN:
					{
						if (curArg.items.Count != 0)
						{
							throw new UsageException("usage: FORGETLOGIN");
						}

						PurgeSession();
						break;
					}

					case Command.GET:
					{
						if (curArg.items.Count < 1)
						{
							throw new UsageException("usage: GET item [...]");
						}
												
						ArrayList strItemArray = new ArrayList();
						foreach (string strReposItem in curArg.items)
						{
							ValidateReposPath(strReposItem);
							strItemArray.Add(strReposItem);
						}

						bSuccess = ProcessCommandGet(strItemArray);
						break;
					}

					case Command.GETLABEL:
					{
						if(curArg.items.Count < 2 || curArg.items.Count > 3)
						{
							throw new UsageException("usage: GETLABEL repository_item label [labelpath]");
						}

						string strReposItem = (string) curArg.items[0];
						ValidateReposPath(strReposItem);
						
						string strLabel = (string) curArg.items[1];

						string strLabelPath = (curArg.items.Count == 3) ? (string)curArg.items[2] : null;

						bSuccess = ProcessCommandGetLabel(strReposItem, strLabel, strLabelPath);
						break;
					}
					case Command.GETLABELDIFFS:
					{
						if(curArg.items.Count < 2 || curArg.items.Count > 3)
						{
							throw new UsageException("usage: GETLABELDIFFS repository_item label1 [label2]");
						}

						string strReposPath = (string) curArg.items[0];
						ValidateReposPath(strReposPath);

						string strLabel1 = (string) curArg.items[1];
						string strLabel2 = "";

						if ( curArg.items.Count > 2 )
							strLabel2 = (string)curArg.items[2];

						bSuccess = ProcessCommandGetLabelDiffs(strReposPath, strLabel1, strLabel2);
						break;
					}
					case Command.GETLABELLEDFILES:
					{
						if(curArg.items.Count < 2)
						{
							throw new UsageException("usage: GETLABELLEDFILES repospath label [...]");
						}

						string strReposItem = (string) curArg.items[0];
						ValidateReposPath(strReposItem);

						string strLabel = (string) curArg.items[1];

						bSuccess = ProcessCommandGetLabelledFiles(strReposItem, strLabel);
						break;
					}
					case Command.GETVERSION:
					{
						if (curArg.items.Count != 3)
						{
							throw new UsageException("usage: GETVERSION version item destination_folder");
						}

						int version = Int32.Parse((string) curArg.items[0]);
						string strReposItem = (string) curArg.items[1];
						ValidateReposPath(strReposItem);
						string strDestFolder = (string) curArg.items[2];
						if (!Directory.Exists(strDestFolder))
						{
							Directory.CreateDirectory(strDestFolder);
						}
						if (!Directory.Exists(strDestFolder))
						{
							throw new Exception(string.Format("{0} does not exist and could not be created", strDestFolder));
						}

						bSuccess = ProcessCommandGetVersion(version, strReposItem, strDestFolder);
						break;
					}

					case Command.GETWILDCARD:
					{
						if(curArg.items.Count < 2)
						{
							throw new UsageException("usage: GETWILDCARD repospath wildcard [...]");
						}

						string strReposItem = (string) curArg.items[0];
						ValidateReposPath(strReposItem);

						ArrayList strWildcardArray = new ArrayList();
						for(int i = 1; i < curArg.items.Count; i++)
						{
							strWildcardArray.Add((string) curArg.items[i]);
						}

						bSuccess = ProcessCommandGetWildcard(strReposItem, strWildcardArray);
						break;
					}

					case Command.HELP:
					{
						Help help = new Help(_xml);

						if(curArg.items.Count < 1)
							help.Write();
						else
							help.Write((string) curArg.items[0]);

						break;
					}

					case Command.HISTORY:
					{
						if (curArg.items.Count != 1)
						{
							throw new UsageException("usage: HISTORY repository_path");
						}

						string strReposPath = (string) curArg.items[0];
						ValidateReposPath(strReposPath);

						ProcessCommandHistory(strReposPath);
						break;
					}

					case Command.LABEL:
					{
						if(curArg.items.Count != 2)
						{
							throw new UsageException("usage: LABEL repository_item labelname");
						}

						string strReposPath = (string) curArg.items[0];
						ValidateReposPath(strReposPath);

						string labelName = (string) curArg.items[1];

						ProcessCommandLabel(strReposPath, labelName, VaultDefine.Latest);
						break;
					}

					case Command.LISTALLBRANCHPOINTS:
					{
						if (curArg.items.Count != 0)
						{
							throw new UsageException("usage: LISTALLBRANCHPOINTS");
						}

						bSuccess = ProcessCommandListAllBranchPoints();
						break;
					}

					case Command.LISTCHANGESET:
					{
						if (curArg.items.Count != 0)
						{
							throw new UsageException("usage: LISTCHANGESET");
						}

						bSuccess = ProcessCommandListChangeSet();
						break;
					}

					case Command.LISTCHECKOUTS:
					{
						if (curArg.items.Count != 0)
						{
							throw new UsageException("usage: LISTCHECKOUTS");
						}

						bSuccess = ProcessCommandListCheckOuts();
						break;
					}

					case Command.LISTFOLDER:
					{
						if (curArg.items.Count != 1)
						{
							throw new UsageException("usage: LISTFOLDER repository_folder");
						}

						string strReposFolder = (string) curArg.items[0];
						ValidateReposPath(strReposFolder);

						bSuccess = ProcessCommandListFolder(strReposFolder);
						break;
					}

					case Command.LISTREPOSITORIES:
					{
						if (curArg.items.Count != 0)
						{
							throw new UsageException("usage: LISTREPOSITORIES");
						}

						bSuccess = ProcessCommandListRepositories();
						break;
					}

					case Command.LISTUSERS:
					{
						if (curArg.items.Count != 0)
						{
							throw new UsageException("usage: LISTUSERS");
						}

						Login();
						DoListUsers();

						break;
					}

					case Command.LISTWORKINGFOLDERS:
					{
						if (curArg.items.Count != 0)
						{
							throw new UsageException("usage: LISTWORKINGFOLDERS");
						}

						Login();	
						WriteListWorkingFolders();

						break;
					}

					case Command.MOVE:
					{
						if (curArg.items.Count != 2)
						{
							throw new UsageException("usage: MOVE path_from path_to");
						}

						string strReposPath_From = (string) curArg.items[0];
						ValidateReposPath(strReposPath_From);

						string strReposPath_To = (string) curArg.items[1];
						ValidateReposPath(strReposPath_To);

						bSuccess = ProcessCommandMove(strReposPath_From, strReposPath_To);
						break;
					}

					case Command.REMEMBERLOGIN:
					{
						if (curArg.items.Count != 0)
						{
							throw new UsageException("This command accepts no arguments.");
						}

						Login(false, true);
						break;
					}

					case Command.PIN:
					{
						if (curArg.items.Count != 2)
						{
							throw new UsageException("usage: PIN repository_path version");
						}

						string strReposPath = RepositoryPath.NormalizeFolder((string) curArg.items[0]);
						ValidateReposPath(strReposPath);

						string strVersion = (string) curArg.items[1];
						int version = int.Parse(strVersion);

						bSuccess = ProcessCommandPin(strReposPath, version);
						break;
					}

					case Command.RENAME:
					{
						if (curArg.items.Count != 2)
						{
							throw new UsageException("usage: RENAME from_name to_name");
						}

						string strReposPath = RepositoryPath.NormalizeFolder((string) curArg.items[0]);
						ValidateReposPath(strReposPath);

						string strNewName = (string) curArg.items[1];

						bSuccess = ProcessCommandRename(strReposPath, strNewName);
						break;
					}

					case Command.SETLABELWORKINGFOLDER:
					{
						if(curArg.items.Count < 3 || curArg.items.Count > 4)
						{
							throw new UsageException("usage: SETLABELWORKINGFOLDER repository_item label local_folder [labelpath]");
						}

						string strReposFolder = RepositoryPath.NormalizeFolder((string) curArg.items[0]);
						ValidateReposPath(strReposFolder);

						string strLabel = (string)curArg.items[1];
						string strLocalFolder = (string) curArg.items[2];

						string strLabelSubItem = (curArg.items.Count == 4) ? (string) curArg.items[3] : null;

						bSuccess = ProcessCommandSetLabelWorkingFolder(strReposFolder, strLabel, strLocalFolder, strLabelSubItem);
						break;
					}

					case Command.SETWORKINGFOLDER:
					{
						if (curArg.items.Count != 2)
						{
							throw new UsageException("usage: SETWORKINGFOLDER repository_folder local_folder");
						}

						string strReposFolder = RepositoryPath.NormalizeFolder((string) curArg.items[0]);
						ValidateReposPath(strReposFolder);

						string strDiskFolder = (string) curArg.items[1];
						if (!Directory.Exists(strDiskFolder))
						{
							Directory.CreateDirectory(strDiskFolder);
						}
						if (!Directory.Exists(strDiskFolder))
						{
							throw new Exception(string.Format("{0} does not exist and could not be created", strDiskFolder));
						}

						Login();	

						_ci.TreeCache.SetWorkingFolder(strReposFolder, strDiskFolder);
						
						WriteListWorkingFolders();

						break;
					}

					case Command.SHARE:
					{
						if (curArg.items.Count != 2)
						{
							throw new UsageException("usage: SHARE repository_path_from repository_path_to");
						}

						string strReposPath_From = (string) curArg.items[0];
						ValidateReposPath(strReposPath_From);

						string strReposPath_To = (string) curArg.items[1];
						ValidateReposPath(strReposPath_To);

						bSuccess = ProcessCommandShare(strReposPath_From, strReposPath_To);
						break;
					}

					case Command.UNCLOAK:
					{
						if (curArg.items.Count < 1)
						{
							throw new UsageException("usage: CLOAK item [...]");
						}

						ArrayList strItemArray = new ArrayList();
						foreach (string strReposItem in curArg.items)
						{
							ValidateReposPath(strReposItem);
							strItemArray.Add(strReposItem);
						}

						bSuccess = ProcessCommandUncloak(strItemArray);
						break;
					}

					case Command.UNDOCHANGESETITEM:
					{
						if ( curArg.items.Count != 1 )
						{
							throw new UsageException("usage: UNDOCHANGESETITEM changesetitem_id");
						}

						int nChgSetItemID = -1;
						try
						{
							nChgSetItemID = Convert.ToInt32(curArg.items[0]);
						}
						catch
						{
							nChgSetItemID = -1;
						}

						bSuccess = ProcessCommandUndoChangeSetItem(nChgSetItemID);
						break;
					}

					case Command.UNDOCHECKOUT:
					{
						if (curArg.items.Count != 1)
						{
							throw new UsageException("usage: UNDOCHECKOUT repository_path");
						}

						string strReposPath = RepositoryPath.NormalizeFolder((string) curArg.items[0]);
						ValidateReposPath(strReposPath);

						bSuccess = ProcessCommandUndoCheckout(strReposPath);
						break;
					}

					case Command.UNPIN:
					{
						if (curArg.items.Count != 1)
						{
							throw new UsageException("usage: UNPIN repository_path");
						}

						string strReposPath = RepositoryPath.NormalizeFolder((string) curArg.items[0]);
						ValidateReposPath(strReposPath);

						bSuccess = ProcessCommandUnPin(strReposPath);
						break;
					}

					case Command.UNSETWORKINGFOLDER:
					{
						if (curArg.items.Count != 1)
						{
							throw new UsageException("The UNSETWORKINGFOLDER command requires 1 argument.");
						}

						string strReposFolder = (string) curArg.items[0];
						ValidateReposPath(strReposFolder);

						Login();	

						_ci.TreeCache.RemoveWorkingFolder(strReposFolder);

						WriteListWorkingFolders();

						break;
					}

					default:
					{
						throw new UsageException("no command specified - run 'vault.exe HELP' for help");
						//break;
					}
						#endregion
				}
			}
			catch (Exception)
			{
				throw;
			}
			finally
			{
				// force a logout if an error occurred.
				Logout(bSuccess == false);
			}

			return bSuccess;
		}

		private bool PreProcessCommand(Args arg)
		{
			bool bSuccess = true;
			if (arg.Cmd == Command.BATCH)
			{
				if (arg.items.Count == 1)
				{
					ProcessCommandBatch(arg);
				}
				else
				{
					arg.Error = true;
					arg.ErrorMessage = string.Format("usage: {0} file name | -", arg.Cmd.ToString());
				}

				bSuccess = (arg.Error == false);
			}
			return bSuccess;
		}

		private bool ProcessCommandAdd(string strReposFolder, ArrayList strItemArray)
		{
			bool bSuccess = true;

			string strFolderNormalized = RepositoryPath.NormalizeFolder(strReposFolder);
			ValidateReposPath(strFolderNormalized);

			Login();

			_ci.Refresh();

			VaultClientFolder folder = _ci.TreeCache.Repository.Root.FindFolderRecursive(strFolderNormalized);
			if (folder == null)
			{
				throw new Exception(string.Format("{0} does not exist", strFolderNormalized));
			}

			ChangeSetItemColl csic = new ChangeSetItemColl();

			for (int i=0; i < strItemArray.Count; i++)
			{
				string strItem = (string) strItemArray[i];
				string strReposPath = strFolderNormalized + "/" + Path.GetFileName(strItem);

				if (File.Exists(strItem))
				{
					ChangeSetItem_AddFile csaf = new ChangeSetItem_AddFile(
						DateTime.Now, 
						_args.Comment, 
						String.Empty, 
						strItem,
						strReposPath);
					csic.Add(csaf);
				}
				else if (Directory.Exists(strItem))
				{
					ChangeSetItem_AddFolder csaf = new ChangeSetItem_AddFolder(
						DateTime.Now, 
						_args.Comment, 
						String.Empty, 
						strItem,
						strReposPath);
					csic.Add(csaf);
				}
				else
				{
					throw new Exception(string.Format("{0} does not exist", strItem));
				}
			}

			_ci.InternalChangeSet_Append(csic);

			if (_args.AutoCommit)
			{
				_ci.InternalChangeSet_SetComment(_args.Comment);
				WriteChangeSet(csic);
				bSuccess = _ci.Commit(csic);
			}
			else
			{
				WriteChangeSet();
			}
			return bSuccess;
		}

		private bool ProcessCommandAddRepository(string strNewReposName)
		{
			Login();

			_ci.Connection.AddRepository(strNewReposName, true);

			_xml.WriteUserMessage(string.Format("Added repository: {0}", strNewReposName));

			return true;
		}
		
		private bool ProcessCommandAddUser(string strLogin, string strPassword, string strEmail)
		{
			bool bRet = true;
			ArrayList groupList = new ArrayList();

			Login();

			VaultClientNetLib.AdminService.VaultUser newUser = new VaultClientNetLib.AdminService.VaultUser();
			newUser.Login = strLogin;
			newUser.Password = VaultLib.VaultUserCrypt.HashPassword(strLogin, strPassword);
			newUser.Name = strLogin;
			newUser.Email = strEmail;
			newUser.isActive = true;
			newUser.DefaultRights = 7;		
			newUser.BelongToGroups = (VaultClientNetLib.AdminService.VaultGroup[]) groupList.ToArray(typeof(VaultClientNetLib.AdminService.VaultGroup));				

			try
			{
				_ci.Connection.AddUser(ref newUser);
			}
			catch (Exception e)
			{
				string strMsg = null;
				if ( newUser.UserID == 0 )
				{
					strMsg = string.Format("{0} was not created - {1}", strLogin, e.Message);
					bRet = false;
				}
				else
				{
					strMsg = string.Format("{0} was created.  But, there was a small problem - {1}", strLogin, e.Message);
				}
				_xml.WriteUserMessage(strMsg);
			}

			if ( bRet == true )
			{
				DoListUsers();
			}

			return bRet;
		}

		private void ProcessCommandBatch(Args arg)
		{
			arg.SetBatchModeOperation((string)arg.items[0]);
		}

		private bool ProcessCommandBlame(string strReposPath, int linenumber)
		{
			bool bSuccess = true;

			Login();

			VaultClientNetLib.ClientService.VaultBlameNode[] blames = null;
			_ci.Connection.Blame(_ci.ActiveRepositoryID, strReposPath, 0, -1, ref blames);

			foreach (VaultClientNetLib.ClientService.VaultBlameNode bn in blames)
			{
				if (bn.FirstLine <= linenumber && (bn.FirstLine + bn.CountLines - 1) >= linenumber)
				{
					_xml.WriteUserMessage(string.Format("User {0} last changed line {1} of {2} with the comment:\r\n {3}", bn.UserName, linenumber, strReposPath, bn.Comment));
					_xml.Begin("blame");
					_xml.AddPair("user", bn.UserName);
					_xml.AddPair("version", bn.OriginatingVersion);
					_xml.AddPair("comment", bn.Comment);
					_xml.End();
				}
			}

			return bSuccess;
		}

		private bool ProcessCommandBranch(string strReposPath_From, string strReposPath_To)
		{
			bool bSuccess = true;

			Login();

			ChangeSetItemColl csic = new ChangeSetItemColl();

			VaultClientFolder vcfolder = null;
			VaultClientFile vcfile = null;

			vcfolder = _ci.TreeCache.Repository.Root.FindFolderRecursive(strReposPath_From);
			if (vcfolder != null)
			{
				// ok, this is a folder
				ChangeSetItem_CopyBranch csi = new ChangeSetItem_CopyBranch(
					DateTime.Now,
					_args.Comment,
					String.Empty,
					vcfolder.FullPath,
					strReposPath_To,
					vcfolder.ObjVerID);
				csic.Add(csi);
			}
			else
			{
				vcfile = _ci.TreeCache.Repository.Root.FindFileRecursive(strReposPath_From);
				if (vcfile != null)
				{
					throw new UsageException(string.Format("{0} exists, but this command may not be used to branch individual files.", strReposPath_From));
				}
				else
				{
					throw new Exception(string.Format("{0} does not exist", strReposPath_From));
				}
			}

			_ci.InternalChangeSet_Append(csic);

			if (_args.AutoCommit)
			{
				_ci.InternalChangeSet_SetComment(_args.Comment);
				WriteChangeSet(csic);
				bSuccess = _ci.Commit(csic);
			}
			else
			{
				WriteChangeSet();
			}

			return bSuccess;
		}

		private bool ProcessCommandCheckout(ArrayList strItemArray)
		{
			bool bRet = true;

			string strReposItem = null;

			VaultClientFolder vcfolder = null;
			VaultClientFile vcfile = null;

			VaultClientNetLib.ClientService.VaultResponseItem[] vriResponses = null;
			VaultClientNetLib.ClientService.VaultResponseItem vri = null;

			int j, nLen;

			Login();
			_ci.Refresh();

			if(_args.Merge == MergeType.Unspecified)
			{
				// use a default merge type if one was not specified
				_args.Merge = MergeType.AttemptAutomaticMerge;
			}

			for (int i=0; i < strItemArray.Count; i++)
			{
				strReposItem = (string) strItemArray[i];
				vcfolder = _ci.TreeCache.Repository.Root.FindFolderRecursive(strReposItem);
				if (vcfolder != null)
				{
					throw new UsageException(string.Format("{0} exists, but the CHECKOUT command can only be used to checkout individual files", strReposItem));
				}
				else
				{
					vcfile = _ci.TreeCache.Repository.Root.FindFileRecursive(strReposItem);
					if (vcfile != null)
					{
						try
						{
							vriResponses = _ci.CheckOut(vcfile, (_args.CheckOutExclusive ? VaultCheckOutType.Exclusive : VaultCheckOutType.CheckOut), string.Empty);
							if ( vriResponses == null )
							{
								throw new Exception(string.Format("The checkout on {0} did not return a resonse.", strReposItem));
							}
						}
						catch (Exception)
						{
							throw;
						}

						// process the responses
						for (j = 0, nLen = vriResponses.Length; j < nLen; j++)
						{
							vri = vriResponses[j];
							if ( (vri.Status == VaultStatusCode.Success) || 
								(vri.Status == VaultStatusCode.SuccessRequireFileDownload) ) 
							{
								_ci.Get(vcfile, false, MakeWritableType.MakeAllFilesWritable, _args.SetFileTime, _args.Merge, null);
								_xml.WriteUserMessage(string.Format("Checked out: {0}", vcfile.FullPath));
							}
							else
							{
								bRet = false;
								_xml.WriteUserMessage(string.Format("Failed Checkout for {0}: {1}", vcfile.FullPath, VaultConnection.GetSoapExceptionMessage(vri.Status)) );
							}
						}
					}
					else
					{
						throw new Exception(string.Format("{0} does not exist.", strReposItem));
					}
				}
			}

			return bRet;
		}

		private bool ProcessCommandCloak(ArrayList strItemArray)
		{
			Login();
			_ci.Refresh();

			for (int i=0; i < strItemArray.Count; i++)
			{
				string strReposItem = (string) strItemArray[i];

				VaultClientFolder vcfolder = null;

				vcfolder = _ci.TreeCache.Repository.Root.FindFolderRecursive(strReposItem);
				if (vcfolder == null)
				{
					throw new UsageException(string.Format("{0} exists, but the CLOAK command can only be used on folders.", strReposItem));
				}
				else
				{
					_ci.CloakOrUnCloak(vcfolder.FullPath, true);
					_xml.WriteUserMessage(string.Format("Cloaked: {0}", vcfolder.FullPath));
				}
			}

			return true;
		}

		private bool ProcessCommandCommit(ArrayList strItemArray)
		{
			ChangeSetItem csiItem = null;
			bool bRet = false;

			// login to the server.
			Login();

			if ( _ci.WorkingFolderOptions.RequireCheckOutBeforeCheckIn == false )
			{
				// do a scan to update the change set list
				_ci.UpdateKnownChanges_All(false);
			}

			// get the internal change set
			ChangeSetItemColl csic = _ci.InternalChangeSet_GetItems(true);
			if ( (csic != null) && (csic.Count > 0) )
			{
				// a sub set of the change set is requested... build that collection
				if ( strItemArray.Count > 0 )
				{
					int nPos = 0;

					// set the old change set list.
					ChangeSetItemColl csicOld = csic;

					// the new list of change set items
					csic = new ChangeSetItemColl();
					
					// find the subset of change sets to use
					for (int i = 0; i < strItemArray.Count; i++)
					{
						string strReposItem = (string) strItemArray[i];

						// see if the item is numeric based
						try
						{
							nPos = Convert.ToInt32(strReposItem);
						}
						catch
						{
							nPos = -1;
						}

						if ( nPos == -1 )
						{
							// a string based subset item
							ValidateReposPath(strReposItem);
							
							// find this item in the old change set
							bool bFoundItem = false;
							for (int j = 0; j < csicOld.Count; j++)
							{
								csiItem = csicOld[j];
								if ( string.Compare( csiItem.DisplayRepositoryPath, strReposItem, true) == 0 )
								{
									// if not already there, add the item
									if ( csic.Contains(csiItem) == false )
									{
										csic.Add(csiItem);
									}

									bFoundItem = true;
									break;
								}
							}

							if ( bFoundItem == false )
							{
								// throw here
								throw new Exception(string.Format("{0} does not exist in the current change set.", strReposItem));
							}
						}
						else 
						{
							if ( ValidateChangeSetItemID(nPos, csicOld) == true )
							{
								csiItem = csicOld[nPos];
								// if not already there, add the item
								if ( csic.Contains(csiItem) == false )
								{
									csic.Add(csiItem);
								}
							}
							else
							{
								throw new UsageException(string.Format("Invalid ChangeSetItem ID: {0}.  Please use the LISTCHANGESET command to retrieve a valid ID.", nPos));
							}
						}
					}
				}


				// of the finalized change set, locate files which need to be "un-checked out"
				VaultClientFileColl vcfcUndoCheckouts = new VaultClientFileColl();
				VaultClientFile vcfile = null;
				ChangeSetItemColl csiRemove = new ChangeSetItemColl();

				for (int i = 0; i < csic.Count; i++)
				{
					csiItem = csic[i];
					if ( csiItem is ChangeSetItem_Unmodified )
					{
						vcfile = _ci.TreeCache.Repository.Root.FindFileRecursive(csiItem.DisplayRepositoryPath);
						if ( 
							(vcfile != null) && 
							( ((ChangeSetItem_Unmodified)csiItem).FileID == vcfile.ID )
							)
						{
							// this file is unmodified, what do we do about it?
							// default is to leave checked out

							// maybe we checkin unchanged files
							if (_args.Unchanged == UnchangedHandler.LeaveCheckedOut)
							{
								// this change set item needs to be 
								// removed from the items to commit
								// since it should be left checked out.
								csiRemove.Add(csiItem);
							}
							else if (_args.Unchanged == UnchangedHandler.UndoCheckout) 
							{
								// regardless of .KeepChecked out, this 
								// change set item needs to be removed 
								// from the items to commit
								csiRemove.Add(csiItem);

								// when keep checked out has not been specified...
								if ( _args.KeepCheckedOut == false )
								{
									// ...we undo checkouts
									if(_ci.GetWorkingFolder(vcfile.Parent) == null)
									{
										throw new Exception(string.Format("{0} does not have a working folder set", vcfile.Name));
									}
									vcfcUndoCheckouts.Add(vcfile);
								}
							}
						}
						else
						{
							throw new Exception(String.Format("{0} does not exist", csiItem.DisplayRepositoryPath));
						}
					}
				}

				// set the comment
				_ci.InternalChangeSet_SetComment(_args.Comment);

				// remove any change set items which will not be committed
				// with the changeset.
				for (int i = 0; i < csiRemove.Count; i++)
				{
					csic.Remove( csiRemove[i] );
				}

				// write out the change set after the items have been removed.
				WriteChangeSet(csic);

				// commit the transaction
				bRet = _ci.Commit(csic, _args.KeepCheckedOut, false);
				if ( (bRet == true) && (vcfcUndoCheckouts.Count > 0) )
				{
					// the commit was successful, now undo checkouts.
					_ci.UndoCheckOut( (VaultClientFile[]) vcfcUndoCheckouts.ToArray( typeof(VaultClientFile) ), _args.LocalCopy);
				}
			}
			else
			{
				// nothing to do but... 
				bRet = true;

				// write out the change set
				WriteChangeSet(csic);

			}

			return bRet;
		}

		private bool ProcessCommandCreateFolder(string strReposFolder)
		{
			bool bSuccess = true;

			Login();

			ChangeSetItem_CreateFolder csaf = new ChangeSetItem_CreateFolder(
				DateTime.Now, 
				_args.Comment, 
				String.Empty, 
				strReposFolder);
			ChangeSetItemColl csic = new ChangeSetItemColl();
			csic.Add(csaf);

			_ci.InternalChangeSet_Append(csic);

			if (_args.AutoCommit)
			{
				_ci.InternalChangeSet_SetComment(_args.Comment);

				WriteChangeSet(csic);
				bSuccess = _ci.Commit(csic);
			}
			else
			{
				WriteChangeSet();
			}

			return bSuccess;
		}

		private bool ProcessCommandDelete(ArrayList strItemArray)
		{
			bool bSuccess = true;
			
			Login();
			_ci.Refresh();

			ChangeSetItemColl csic = new ChangeSetItemColl();

			for (int i=0; i < strItemArray.Count; i++)
			{
				string strReposItem = (string) strItemArray[i];

				VaultClientFolder vcfolder = null;
				VaultClientFile vcfile = null;

				vcfolder = _ci.TreeCache.Repository.Root.FindFolderRecursive(strReposItem);
				if (vcfolder != null)
				{
					// OK, this is a folder
					ChangeSetItem_DeleteFolder csdf = new ChangeSetItem_DeleteFolder(
						DateTime.Now,
						_args.Comment,
						String.Empty,
						vcfolder.ID,
						vcfolder.FullPath);
					csic.Add(csdf);
				}
				else
				{
					vcfile = _ci.TreeCache.Repository.Root.FindFileRecursive(strReposItem);
					if (vcfile != null)
					{
						// OK, this is a file
						ChangeSetItem_DeleteFile csdf = new ChangeSetItem_DeleteFile(
							DateTime.Now,
							_args.Comment,
							String.Empty,
							vcfile.ID,
							vcfile.FullPath);
						csic.Add(csdf);
					}
					else
					{
						throw new Exception(string.Format("{0} does not exist", strReposItem));
					}
				}
			}

			_ci.InternalChangeSet_Append(csic);

			if (_args.AutoCommit)
			{
				_ci.InternalChangeSet_SetComment(_args.Comment);

				WriteChangeSet(csic);
				bSuccess = _ci.Commit(csic);
			}
			else
			{
				WriteChangeSet();
			}

			return bSuccess;
		}

		private bool ProcessCommandDiff 
		( 
			string strArgDiffBin, 
			string strArgDiffArgs, 
			DiffAgainstType datDiffChoice, 
			bool bRecursive, 
			string strLeft, 
			string strLeftLbl, 
			string strRight, 
			string strRightLbl
		)
		{
			int nDoDiffError = DoDiffError.Success;

			// login to the server.
			Login();
			_ci.Refresh();

			// determine the diff command
			string strDiffBin = strArgDiffBin;
			if ( (strDiffBin == null) || (strDiffBin.Length == 0) )
			{
				// nothing specified on the command line, what about the environment variable.
				strDiffBin = Environment.GetEnvironmentVariable(VaultCmdLineClientDefines.DiffEnv);
				
				if ( strDiffBin == null )
				{
					// just try plain diff
					strDiffBin = VaultCmdLineClientDefines.DiffBin;
				}
			}

			// add parameters for left/right to the args
			string strFullDiffArgs = strArgDiffArgs;
			if ( strFullDiffArgs.IndexOf(VaultCmdLineClientDefines.DiffLeftItem) < 0 )
			{
				strFullDiffArgs += string.Format("\"{0}\" ", VaultCmdLineClientDefines.DiffLeftItem);
			}
			if ( strFullDiffArgs.IndexOf(VaultCmdLineClientDefines.DiffRightItem) < 0 )
			{
				strFullDiffArgs += string.Format("\"{0}\" ", VaultCmdLineClientDefines.DiffRightItem);
			}

			string strLItem = null, strRItem = null;
			Exception e = null;

			// get the repository item.
			VaultClientTreeObject to = _ci.TreeCache.Repository.Root.FindTreeObjectRecursive(strLeft);
			if ( to != null )
			{
				if ( to is VaultClientFile )
				{
					VaultClientFile vcfile = (VaultClientFile)to;
					// the working folder must be set
					if ( _ci.TreeCache.GetBestWorkingFolder(vcfile.Parent) != null )
					{
						nDoDiffError = _ci.DoWorkingFileDiff(vcfile, datDiffChoice, strRight, 
							strDiffBin, strFullDiffArgs, strLeftLbl, strRightLbl, out strLItem, out strRItem, out e);
					}
					else
					{
						nDoDiffError = DoDiffError.NoValidWorkingFolder;
					}
				}
				else
				{
					VaultClientFolder vcfolder = (VaultClientFolder)to;
					if ( _ci.TreeCache.GetBestWorkingFolder(vcfolder) != null )
					{
						nDoDiffError = _ci.DoWorkingFolderDiff(vcfolder, bRecursive, datDiffChoice, strRight, 
							strDiffBin, strFullDiffArgs, strLeftLbl, strRightLbl, out strLItem, out strRItem, out e);
					}
					else
					{
						nDoDiffError = DoDiffError.NoValidWorkingFolder;
					}
				}
			}
			else
			{
				// the repository item could not be found.
				nDoDiffError = DoDiffError.RepositoryItemNotFound;
			}

			// handle the diff error
			if ( nDoDiffError != DoDiffError.Success )
			{
				if ( strLeft == null )
				{
					strLeft = "Unknown Item";
				}
				if ( strRight == null )
				{
					strRight = "Unknown Item";
				}

				string strMessage = null;
				switch ( nDoDiffError )
				{
					case DoDiffError.RepositoryItemNotFound:
					case DoDiffError.LeftItemDoesNotExist:
						strMessage = string.Format("Item {0} could not be found.", strLeft);
						break;
					case DoDiffError.RightItemDoesNotExist:
						strMessage = string.Format("Item {0} could not be found.", strRight);
						break;
					case DoDiffError.DiffBinaryError:
						strMessage = "The Diff utility encountered an error during execution.  Please verify the use of VAULTDIFF or the \"diff\" utility.";
						break;
					case DoDiffError.TempPathNotFound:
						strMessage = "Could not find temp path.";
						break;
					case DoDiffError.TempFileNotFound:
						strMessage = "Could not find temp file.";
						break;
					case DoDiffError.LabelNotRetrieved:
						strMessage = string.Format("Could not retrieve label {0}.", strRight);
						break;
					case DoDiffError.NoValidWorkingFolder:
						strMessage = string.Format("The working folder has not been set for {0}.", strLeft);
						break;
					default:
						if ( e != null )
						{
							strMessage = e.Message;
						}
						else
						{
							// some other error
							// TODO - move this to a resource.
							strMessage = "An unknown error occurred executing the diff utility.";
						}
						break;
				}

				throw new Exception(strMessage);
			}

			return (nDoDiffError == DoDiffError.Success);
		}

		private bool ProcessCommandGet(ArrayList strItemArray)
		{
			bool bExists = true;
			bool bSuccess = true;
			bool bResetCloaks = false;
			VaultNameValueCollection cloaks = null;
			string strErrorItem = null;
			
			// get the merge option.
			MergeType mt = _args.Merge;

			// this argument specifies a get to a non-working folder
			// in this case, there are only two valid options.
			if (_args.DestPath != null)
			{
				// on a non working folder only valid options :
				//	a) overwrite OR b) do not overwrite (later)
				switch (mt)
				{
					case MergeType.AttemptAutomaticMerge:
						// in this case, automatic merge is not possible, switch to do not overwrite
						mt = MergeType.MergeLater;
						break;
					case MergeType.OverwriteWorkingCopy:
					case MergeType.MergeLater:
						// do nothing - set correctly
						break;
					default:
						// the default value
						mt = MergeType.OverwriteWorkingCopy;
						break;
				}
			}
			if(mt == MergeType.Unspecified)
			{
				// use a default merge type if a valid merge type was not set
				mt = MergeType.AttemptAutomaticMerge;
			}

			// login to the server.
			Login();
			_ci.Refresh();

			// after login - check for flag to use/not use cloak 
			if(_args.RespectCloaks == false)
			{
				cloaks = _ci.TreeCache.Cloaks;
				bResetCloaks = true;
				_ci.TreeCache.Cloaks = null;
			}

			for (int i=0; i < strItemArray.Count; i++)
			{
				string strReposItem = (string) strItemArray[i];

				VaultClientFile vcfile = null;
				VaultClientFolder vcfolder = _ci.TreeCache.Repository.Root.FindFolderRecursive(strReposItem);
				if (vcfolder != null)
				{
					// OK, this is a folder
					if (_args.DestPath != null)
					{
						_ci.GetToNonWorkingFolder(vcfolder, _args.Recursive, true, (mt == MergeType.OverwriteWorkingCopy) ? true : false, _args.MakeWritable, _args.SetFileTime, _args.DestPath, null);
					}
					else
					{
						// a get to a working folder
						_ci.Get(vcfolder, _args.Recursive, true, _args.MakeWritable, _args.SetFileTime, mt, null);
					}
				}
				else
				{
					vcfile = _ci.TreeCache.Repository.Root.FindFileRecursive(strReposItem);
					if (vcfile != null)
					{
						// OK, this is a file
						if (_args.DestPath != null)
						{
							_ci.GetToNonWorkingFolder(vcfile, true, (mt == MergeType.OverwriteWorkingCopy) ? true : false, _args.MakeWritable, _args.SetFileTime, vcfile.Parent.FullPath, _args.DestPath, null);
						}
						else
						{
							_ci.Get(vcfile, true, _args.MakeWritable, _args.SetFileTime, mt, null);
						}
					}
					else
					{
						bExists = false;
						strErrorItem = strReposItem;
					}
				}
			}

			if ( 
				(_args.RespectCloaks == false) && 
				(bResetCloaks == true) 
				)
			{
				_ci.TreeCache.Cloaks = cloaks;
			}

			if(bExists == false)
				throw new Exception(string.Format("{0} does not exist", strErrorItem));

			return bSuccess;
		}

		private bool ProcessCommandGetLabel(string reposItem, string label, string labelSubItem)
		{
			bool bSuccess = true;
			string[] discoveredPaths;
			VaultClientTreeObject labelStructure;
			long labelSubItemId = 0;
			long labelID = 0;

			// retrieve the merge option.
			MergeType mt = _args.Merge;
			// TODO - SEE IF YOU CAN ACTUALLY SET AUTOMATIC MERGE FOR A LABEL GET
			//			if(_args.DestPath != null)
			//			{
			// a get to a non-working folder
			switch (mt)
			{
				case MergeType.AttemptAutomaticMerge:
					// in this case, automatic merge is not possible, switch to do not overwrite
					mt = MergeType.MergeLater;
					break;
				case MergeType.OverwriteWorkingCopy:
				case MergeType.MergeLater:
					// do nothing - set correctly
					break;
				default:
					// the default value
					mt = MergeType.OverwriteWorkingCopy;
					break;
			}
			//			}
			//			if ( mt == MergeType.Unspecified )
			//			{
			//				// use a default merge type if a valid merge type was not set
			//				mt = MergeType.AttemptAutomaticMerge;
			//			}

			Login();
			_ci.Refresh();

			VaultClientTreeObject reposTreeObj = _ci.TreeCache.Repository.Root.FindTreeObjectRecursive(reposItem);

			if(reposTreeObj == null)
			{
				throw new Exception("The specified item was not found in the repository.");
			}

			labelSubItemId = reposTreeObj.ID;

			// This is the ID of the file/folder where the label structure was
			// created.
			long rootID = 0;

			bSuccess = _ci.GetByLabel_GetStructure(reposItem, label, ref labelID, labelSubItem, out discoveredPaths, out labelStructure, out rootID);

			if(bSuccess == true)
			{
				VaultClientTreeObjectColl treeObjects = new VaultClientTreeObjectColl();

				if (labelStructure is VaultClientFile)
				{
					if (((VaultClientFile) labelStructure).ID == labelSubItemId)
						treeObjects.Add(labelStructure);
				}
				else
				{
					((VaultClientFolder) labelStructure).FindTreeObjectsRecursive(labelSubItemId, ref treeObjects);
				}

				if(treeObjects.Count < 1)
				{
					throw new Exception("The specified item could not be found in the label.");
				}
				else if(treeObjects.Count > 1)
				{
					throw new Exception("The specified item was not specific within the label.");
				}

				VaultClientTreeObject treeObject = (VaultClientTreeObject)treeObjects[0];
				labelSubItem = discoveredPaths[0];

				// get to non-working folder
				if(_args.DestPath != null)
				{
					if (labelStructure is VaultClientFolder)
					{
						_ci.GetByLabelToNonWorkingFolder_GetData(
							(VaultClientFolder) labelStructure, 
							_args.Recursive, 
							(mt == MergeType.OverwriteWorkingCopy) ? true : false, 
							_args.MakeWritable, 
							_args.SetFileTime, 
							_args.DestPath,
							null, 
							labelID,
							reposItem,
							labelSubItem);
					}
					else
					{
						_ci.GetByLabelToNonWorkingFolder_GetData(
							(VaultClientFile) labelStructure, 
							(mt == MergeType.OverwriteWorkingCopy) ? true : false, 
							_args.MakeWritable, 
							_args.SetFileTime, 
							_args.DestPath, 
							null,
							labelID,
							reposItem,
							labelSubItem);
					}
				}
					// get to working folder
				else
				{
					VaultClientFolder labelRootFolder;

					if(treeObject == null)
					{
						throw new Exception(string.Format("{0} does not exist in the label structure for {1}", reposItem, label));
					}

					if(labelStructure is VaultClientFolder)
					{				
						labelRootFolder = (VaultClientFolder)treeObjects[0];
					}
					else
					{
						labelRootFolder = GetFakeLabelParent((VaultClientFile) labelStructure, reposItem, label);
					}

					string workingFolder = _ci.TreeCache.GetBestWorkingFolder(labelRootFolder);
					if(workingFolder == null)
					{
						throw new Exception(string.Format("Use SETLABELWORKINGFOLDER to set a working folder for {0} in label {1}, or use the -destpath option to get to a non-working folder.", reposItem, label));
					}

					Console.Write("working folder: {0}\n", workingFolder);

					if(labelStructure is VaultClientFolder)
					{
						_ci.GetByLabel_GetData((VaultClientFolder) labelStructure,
							_args.Recursive,
							_args.MakeWritable,
							_args.SetFileTime,
							mt,
							null,
							labelID,
							reposItem,
							labelSubItem);
					}
					else
					{
						_ci.GetByLabel_GetData((VaultClientFile) labelStructure,
							_args.MakeWritable,
							_args.SetFileTime,
							mt,
							null,
							labelID,
							reposItem,
							labelSubItem);
					}
				}
			}
			else
			{
				string subItemOptions = String.Empty;
				
				foreach(string item in discoveredPaths)
				{
					subItemOptions += string.Format("   {0}\n", item);
				}

				throw new Exception(string.Format("The specified item is shared to multiple places in the label structure.\n" +
					"Please use \"vault.exe GETLABEL '{0}' '{1}' labelpath\", where labelpath is one of:\n" +
					"\n" +					
					subItemOptions, label, reposItem));
			}

			return bSuccess;
		}

		private bool ProcessCommandGetVersion(int version, string strReposItem, string strDestFolder)
		{
			bool bExists = true;
			bool bSuccess = true;
			bool bResetCloaks = false;
			VaultNameValueCollection cloaks = null;

			// retrieve the merge option.
			MergeType mt = _args.Merge;

			// this always overwrites, unless specified to auto merge
			switch (mt)
			{
				case MergeType.AttemptAutomaticMerge:
					// in this case, automatic merge is not possible, switch to do not overwrite
					mt = MergeType.MergeLater;
					break;
				case MergeType.OverwriteWorkingCopy:
				case MergeType.MergeLater:
					// do nothing - set correctly
					break;
				default:
					// the default value
					mt = MergeType.OverwriteWorkingCopy;
					break;
			}
			
			Login();
			_ci.Refresh();

			VaultClientFolder vcfolder = null;
			VaultClientFile vcfile = null;

			// temporarily blow away cloaks if that's what the user wanted
			// do after login
			if(_args.RespectCloaks == false)
			{
				cloaks = _ci.TreeCache.Cloaks;
				bResetCloaks = true;
				_ci.TreeCache.Cloaks = null;
			}

			vcfolder = _ci.TreeCache.Repository.Root.FindFolderRecursive(strReposItem);

			if(strDestFolder == null)
				throw new Exception("destination folder is null, you can only checkout historical versions to a non-working folder");

			if (vcfolder != null)
			{
				vcfolder.Version = version;

				VaultClientNetLib.ClientService.VaultFolderDelta vfDelta = new VaultClientNetLib.ClientService.VaultFolderDelta();
				_ci.Connection.GetBranchStructure(_ci.ActiveRepositoryID, strReposItem, vcfolder.ID, version, ref vfDelta, false);

				vcfolder = new VaultClientFolder(vfDelta, vcfolder.Parent);

				_ci.GetByDisplayVersionToNonWorkingFolder(vcfolder, _args.Recursive, _args.MakeWritable, _args.SetFileTime, strDestFolder, null);				
			}
			else
			{
				vcfile = new VaultClientFile(_ci.TreeCache.Repository.Root.FindFileRecursive(strReposItem));
				if (vcfile != null)
				{
					// Set the version.
					vcfile.Version = version;
					_ci.GetByDisplayVersionToNonWorkingFolder(vcfile, _args.MakeWritable, _args.SetFileTime, vcfile.Parent.FullPath, strDestFolder, null);
				}
				else
				{
					bExists = false;
				}
			}

			if (
				(_args.RespectCloaks == false) && 
				(bResetCloaks == true) 
				)
			{
				_ci.TreeCache.Cloaks = cloaks;
			}

			if(bExists == false)
				throw new Exception(string.Format("{0} does not exist", strReposItem));

			return bSuccess;
		}

		
		private bool ProcessCommandGetLabelledFiles(string strReposPath, string strLabel)
		{
			bool bSuccess = true;
			bool bResetCloaks = false;
			VaultNameValueCollection cloaks = null;
			ArrayList RegexArray = new ArrayList();

			// get the merge option.
			MergeType mt = _args.Merge;
		
			// this argument specifies a get to a non-working folder
			// in this case, there are only two valid options.
			if (_args.DestPath != null)
			{
				// on a non working folder only valid options :
				//	a) overwrite OR b) do not overwrite (later)
				switch (mt)
				{
					case MergeType.AttemptAutomaticMerge:
						// in this case, automatic merge is not possible, switch to do not overwrite
						mt = MergeType.MergeLater;
						break;
					case MergeType.OverwriteWorkingCopy:
					case MergeType.MergeLater:
						// do nothing - set correctly
						break;
					default:
						// the default value
						mt = MergeType.OverwriteWorkingCopy;
						break;
				}
			}
			if(mt == MergeType.Unspecified)
			{
				// use a default merge type if a valid merge type was not set
				mt = MergeType.AttemptAutomaticMerge;
			}
		
			Login();
			_ci.Refresh();

			// reset cloaks - must come after login
			if(_args.RespectCloaks == false)
			{
				cloaks = _ci.TreeCache.Cloaks;
				bResetCloaks = true;
				_ci.TreeCache.Cloaks = null;
			}

			//OperatingSystem os = Environment.OSVersion;
			//bool bCaseInsensitive = ( (os.Platform == PlatformID.Win32NT) || 
			//									(os.Platform == PlatformID.Win32S) || 
			//									(os.Platform == PlatformID.Win32Windows) ||
			//									(os.Platform == PlatformID.WinCE) );
		
			// for consistency's sake, currently doing case insensitve regex for wildcard
			//			bool bCaseInsensitive = true; 

			VaultClientFolder vcfolder = _ci.TreeCache.Repository.Root.FindFolderRecursive(strReposPath);

			if(vcfolder != null)
			{
				// Get a collection of files below the source folder specified

				VaultClientOperationsLib.VaultClientFileColl vcFileColl = new 
					VaultClientOperationsLib.VaultClientFileColl();
				vcfolder.GetFilesRecursive(ref vcFileColl,null);

				string destPath = _args.DestPath;
				string myWorkingFolder;

				// Loop through the files and try to get a specific labelled version of each one.
				foreach (VaultClientOperationsLib.VaultClientFile vcFile in vcFileColl)
				{
					try
					{
						
						// If the destfolder argument was provided then create the full
						// path on the local disk for each file.
						if (destPath.Length != 0)
						{
							myWorkingFolder = destPath + vcFile.Parent.FullPath.Replace(strReposPath,"").Replace("/",@"\");
						}
							// Otherwise, just use the actual working folder for this user.
						else
						{
							WorkingFolder wf = _ci.GetWorkingFolder(vcFile);
							myWorkingFolder = wf.GetLocalFolderPath();
						}

						// Use the Vault libraries built in function for getting a specific
						// labelled version of a file.
						_ci.GetLabeledFileToNonWorkingFolder(vcFile,
							strLabel,
							null,
							ref myWorkingFolder );

						if ( _args.Verbose )
							Console.WriteLine("Getting file: " + vcFile.FullPath);

					}
					catch(Exception)
					{
						// If the label does not exist, the library function will actually
						// throw an error.  So we will not output any error message.
						//Console.WriteLine(ex.Message);
					}
														
				}

			}

			if (
				(_args.RespectCloaks == false) && 
				(bResetCloaks == true) 
				)
			{
				_ci.TreeCache.Cloaks = cloaks;
			}

			if(vcfolder == null && _ci.TreeCache.Repository.Root.FindFileRecursive(strReposPath) != null)
				throw new Exception(string.Format("{0} is a file, not a directory", strReposPath));
			else if(vcfolder == null)
				throw new Exception(string.Format("{0} does not exist", strReposPath));

			return bSuccess;
		}

		private bool ProcessCommandGetLabelDiffs(string strReposPath, string strLabel1, string strLabel2)
		{
			bool bSuccess = true;

			Login();

			VaultHistoryQueryRequest hq = new VaultHistoryQueryRequest();
			hq.Recursive = true;
			
			DateTime beginDate = VaultDate.EmptyDate(), endDate = VaultDate.EmptyDate();
			GetLabelTimeStamps(strReposPath, strLabel1, strLabel2, out beginDate, out endDate);

			hq.BeginDate = beginDate;

			if ( endDate != VaultDate.EmptyDate() )
				hq.EndDate = endDate;
			else hq.EndDate = System.DateTime.Now;
			
			hq.RepID = _ci.ActiveRepositoryID;

			VaultClientFolder vcfolder = null;
			VaultClientFile vcfile = null;

			vcfolder = _ci.TreeCache.Repository.Root.FindFolderRecursive(strReposPath);
			if (vcfolder != null)
			{
				hq.TopName = vcfolder.FullPath;
				hq.TopID = vcfolder.ID;
				hq.IsFolder = true;
			}
			else
			{
				vcfile = _ci.TreeCache.Repository.Root.FindFileRecursive(strReposPath);
				if (vcfile != null)
				{
					hq.TopName = vcfile.FullPath;
					hq.TopID = vcfile.ID;
					hq.IsFolder = false;
				}
				else
				{
					throw new Exception(string.Format("{0} does not exist", strReposPath));
				}
			}

			int nRowsRetrieved = 0;
			string strQryToken = null;

			_ci.Connection.HistoryBegin(hq, _args.HistoryRowLimit, ref nRowsRetrieved, ref strQryToken);
			VaultHistoryItem[] histitems = null;
			_ci.Connection.HistoryFetch(strQryToken, 0, nRowsRetrieved-1, ref histitems);
			_ci.Connection.HistoryEnd(strQryToken);

			_xml.Begin("history");

			// write header
			_xml.AddPair("fromLabel", strLabel1);
			_xml.AddPair("fromDate", beginDate.ToString());
			if ( strLabel2.Length > 0 )
			{
				_xml.AddPair("toLabel", strLabel2);
				_xml.AddPair("toDate", endDate.ToString());
			}
			else
			{
				_xml.AddPair("toLabel", "");
				_xml.AddPair("toDate", "");
			}

			// sort history items before outputting as XML
			Array.Sort(histitems, new ReverseHistoryItemComparer());		

			// dump history items to XML stream
			foreach (VaultHistoryItem hi in histitems)
			{
				// skip label history - not interested in that.
				if ( hi.HistItemType == VaultHistoryType.Label )
					continue;
				_xml.Begin("item");
				_xml.AddPair("txid", hi.TxID);
				_xml.AddPair("date", hi.TxDate.ToString());
				_xml.AddPair("name", hi.Name);
				_xml.AddPair("version", hi.Version);
				_xml.AddPair("user", hi.UserLogin);
				_xml.AddPair("action", TranslateActionToString(hi) );
				if (
					(hi.Comment != null)
					&& (hi.Comment.Length > 0)
					)
				{
					_xml.AddPair("comment", hi.Comment);
				}
				_xml.End();
			}
			_xml.End();
			return bSuccess;
		}

		private bool ProcessCommandLabel(string strReposPath, string labelName, long versionID)
		{
			bool bSuccess = false;
			int ret = 0;
			VaultClientTreeObject vctreeobj;
			VaultClientNetLib.ClientService.VaultLabelResult labelResult = null;
			long objVerID = -1;

			Login();

			if((vctreeobj = _ci.TreeCache.Repository.Root.FindTreeObjectRecursive(strReposPath)) == null)
			{
				throw new Exception(string.Format("{0} does not exist in the repository.", strReposPath));
			}

			VaultObjectVersionInfo[] ovis = null;
			_ci.Connection.GetObjectVersionList(vctreeobj.ID, ref ovis, false);

			if ( (ovis != null) && (ovis.Length > 0) )
			{
				if ( versionID == VaultDefine.Latest )
				{
					// use the last version, since the list is in ascending order.
					objVerID = ovis[ovis.Length - 1].ObjVerID;
				}
				else
				{
					// find the version number specified for this tree object
					foreach(VaultObjectVersionInfo ovi in ovis)
					{
						if(versionID == ovi.Version)
						{
							objVerID = ovi.ObjVerID;
							break;
						}
					}
				}
			}

			if(objVerID == -1)
			{
				throw new Exception(string.Format("{0} does not exist at version {1}", strReposPath, versionID));
			}


			ret = _ci.AddLabel(strReposPath, objVerID, labelName, _args.Comment, ref labelResult);
			switch(ret)
			{
				case VaultStatusCode.Success:
					bSuccess = true;
					break;

				case VaultStatusCode.FailDuplicateLabel:
					string strErrorMsg = null;
					if(String.Compare(labelResult.ExistingRootPath, strReposPath, true) == 0)
					{
						strErrorMsg = string.Format("{0} already has the label {1} applied", strReposPath, labelName);
					}
					else
					{
						strErrorMsg = string.Format("{0} has inherited the label {1} already", strReposPath, labelName);
					}
					throw new Exception(strErrorMsg);

				default:
					throw new Exception("Label error: " + VaultConnection.GetSoapExceptionMessage(ret));
			}

			return bSuccess;
		}

		private bool ProcessCommandGetWildcard(string strReposPath, ArrayList strWildcardArray)
		{
			bool bSuccess = true;
			bool bResetCloaks = false;
			VaultNameValueCollection cloaks = null;
			ArrayList RegexArray = new ArrayList();

			// get the merge option.
			MergeType mt = _args.Merge;
			
			// this argument specifies a get to a non-working folder
			// in this case, there are only two valid options.
			if (_args.DestPath != null)
			{
				// on a non working folder only valid options :
				//	a) overwrite OR b) do not overwrite (later)
				switch (mt)
				{
					case MergeType.AttemptAutomaticMerge:
						// in this case, automatic merge is not possible, switch to do not overwrite
						mt = MergeType.MergeLater;
						break;
					case MergeType.OverwriteWorkingCopy:
					case MergeType.MergeLater:
						// do nothing - set correctly
						break;
					default:
						// the default value
						mt = MergeType.OverwriteWorkingCopy;
						break;
				}
			}
			if(mt == MergeType.Unspecified)
			{
				// use a default merge type if a valid merge type was not set
				mt = MergeType.AttemptAutomaticMerge;
			}
			
			Login();
			_ci.Refresh();

			// reset cloaks - must come after login
			if(_args.RespectCloaks == false)
			{
				cloaks = _ci.TreeCache.Cloaks;
				bResetCloaks = true;
				_ci.TreeCache.Cloaks = null;
			}

			//OperatingSystem os = Environment.OSVersion;
			//bool bCaseInsensitive = ( (os.Platform == PlatformID.Win32NT) || 
			//									(os.Platform == PlatformID.Win32S) || 
			//									(os.Platform == PlatformID.Win32Windows) ||
			//									(os.Platform == PlatformID.WinCE) );
			
			// for consistency's sake, currently doing case insensitve regex for wildcard
			bool bCaseInsensitive = true; 

			for(int i = 0; i < strWildcardArray.Count; i++)
			{
				Wildcard wildcard = new Wildcard((string) strWildcardArray[i]);
				Regex regex = new Regex(wildcard.ToRegex(bCaseInsensitive));
				RegexArray.Add(regex);
			}
			
			VaultClientFolder vcfolder = _ci.TreeCache.Repository.Root.FindFolderRecursive(strReposPath);

			if(vcfolder != null)
			{
				if(_args.DestPath != null)
				{
					_ci.GetByRegexToNonWorkingFolder(vcfolder, RegexArray, _args.Recursive, true, (mt == MergeType.OverwriteWorkingCopy) ? true : false, _args.MakeWritable, _args.SetFileTime, _args.DestPath, null);
				}
				else
				{
					_ci.GetByRegex(vcfolder, RegexArray, _args.Recursive, true, _args.MakeWritable, _args.SetFileTime, mt, null);
				}
			}

			if (
				(_args.RespectCloaks == false) && 
				(bResetCloaks == true) 
				)
			{
				_ci.TreeCache.Cloaks = cloaks;
			}

			if(vcfolder == null && _ci.TreeCache.Repository.Root.FindFileRecursive(strReposPath) != null)
				throw new Exception(string.Format("{0} is a file, not a directory", strReposPath));
			else if(vcfolder == null)
				throw new Exception(string.Format("{0} does not exist", strReposPath));

			return bSuccess;
		}

		private bool ProcessCommandHistory(string strReposPath)
		{
			bool bSuccess = true;

			Login();

			VaultHistoryQueryRequest hq = new VaultHistoryQueryRequest();
			hq.Recursive = true;

			hq.BeginDate = VaultDate.EmptyDate();
			hq.EndDate = VaultDate.EmptyDate();

			hq.RepID = _ci.ActiveRepositoryID;

			VaultClientFolder vcfolder = null;
			VaultClientFile vcfile = null;

			vcfolder = _ci.TreeCache.Repository.Root.FindFolderRecursive(strReposPath);
			if (vcfolder != null)
			{
				hq.TopName = vcfolder.FullPath;
				hq.TopID = vcfolder.ID;
				hq.IsFolder = true;
			}
			else
			{
				vcfile = _ci.TreeCache.Repository.Root.FindFileRecursive(strReposPath);
				if (vcfile != null)
				{
					hq.TopName = vcfile.FullPath;
					hq.TopID = vcfile.ID;
					hq.IsFolder = false;
				}
				else
				{
					throw new Exception(string.Format("{0} does not exist", strReposPath));
				}
			}

			int nRowsRetrieved = 0;
			string strQryToken = null;

			_ci.Connection.HistoryBegin(hq, _args.HistoryRowLimit, ref nRowsRetrieved, ref strQryToken);
			VaultHistoryItem[] histitems = null;
			_ci.Connection.HistoryFetch(strQryToken, 0, nRowsRetrieved-1, ref histitems);
			_ci.Connection.HistoryEnd(strQryToken);

			_xml.Begin("history");
			foreach (VaultHistoryItem hi in histitems)
			{
				_xml.Begin("item");
				_xml.AddPair("txid", hi.TxID);
				_xml.AddPair("date", hi.TxDate.ToString());
				_xml.AddPair("name", hi.Name);
				_xml.AddPair("type", GetHistItemTypeString(hi.HistItemType));
				_xml.AddPair("version", hi.Version);
				_xml.AddPair("user", hi.UserLogin);
				if (
					(hi.Comment != null)
					&& (hi.Comment.Length > 0)
					)
				{
					_xml.AddPair("comment", hi.Comment);
				}
				_xml.End();
			}
			_xml.End();

			return bSuccess;
		}

		private bool ProcessCommandListAllBranchPoints()
		{
			bool bSuccess = true;
			
			Login();

			VaultBranchPointInfo[] points = null;

			_ci.ListAllBranchPoints(ref points);

			_xml.Begin("listallbranchpoints");
			foreach (VaultBranchPointInfo pt in points)
			{
				if ((pt._objprops & 4) == 4)
				{
					_xml.Begin("branch");
				}
				else
				{
					_xml.Begin("label");
				}
				VaultClientFolder vcfolder_trunk = _ci.TreeCache.Repository.Root.FindFolderRecursive(pt._baselineobjid);
				_xml.AddPair("trunk", vcfolder_trunk.FullPath);
				_xml.AddPair("trunk_versionid", pt._baselineobjverid);

				VaultClientFolder vcfolder_branch = _ci.TreeCache.Repository.Root.FindFolderRecursive(pt._objid);
				_xml.AddPair("branch", vcfolder_branch.FullPath);
				_xml.AddPair("branch_versionid", pt._rootobjverid);

				_xml.AddPair("txid", pt._origintxid);

				_xml.End();
			}
			_xml.End();

			return bSuccess;
		}

		private bool ProcessCommandListChangeSet()
		{
			Login();

			_ci.UpdateKnownChanges_All(true);

			ChangeSetItemColl csic = _ci.InternalChangeSet_GetItems(true);

			WriteChangeSet(csic);

			return true;
		}

		private bool ProcessCommandListCheckOuts()
		{
			Login();

			_ci.Refresh();

			WriteCheckOuts(_ci.TreeCache.CheckOuts);

			return true;
		}

		private bool ProcessCommandListFolder(string strReposFolder)
		{
			Login();

			_ci.Refresh();

			VaultClientFolder vcfolder = _ci.TreeCache.Repository.Root.FindFolderRecursive(strReposFolder);
			if (vcfolder == null)
			{
				throw new Exception(string.Format("{0} does not exist", strReposFolder));
			}

			WriteFolder(vcfolder, _args.Recursive, 0);

			return true;
		}

		private bool ProcessCommandListRepositories()
		{
			Login();

			VaultRepositoryInfo[] reps = null;
			_ci.ListRepositories(ref reps);
					
			_xml.Begin("listrepositories");
			foreach (VaultRepositoryInfo r in reps)
			{
				_xml.Begin("repository");
				_xml.AddPair("name", r.RepName);
				_xml.AddPair("files", r.FileCount);
				_xml.AddPair("folders", r.FolderCount);
				_xml.AddPair("dbsize", r.DbSize);
				_xml.End();
			}
			_xml.End();
					
			return true;
		}

		private bool ProcessCommandMove(string strReposPath_From, string strReposPath_To)
		{
			bool bSuccess = true;
			
			Login();

			ChangeSetItemColl csic = new ChangeSetItemColl();

			VaultClientFolder vcfolder = null;
			VaultClientFile vcfile = null;

			vcfolder = _ci.TreeCache.Repository.Root.FindFolderRecursive(strReposPath_From);
			if (vcfolder != null)
			{
				// ok, this is a folder
				ChangeSetItem_Move csi = new ChangeSetItem_Move(
					DateTime.Now,
					_args.Comment,
					String.Empty,
					vcfolder.ID,
					vcfolder.FullPath,
					strReposPath_To);
				csic.Add(csi);
			}
			else
			{
				vcfile = _ci.TreeCache.Repository.Root.FindFileRecursive(strReposPath_From);
				if (vcfile != null)
				{
					// ok, this is a file
					ChangeSetItem_Move csi = new ChangeSetItem_Move(
						DateTime.Now,
						_args.Comment,
						String.Empty,
						vcfile.ID,
						vcfile.FullPath,
						strReposPath_To);
					csic.Add(csi);
				}
				else
				{
					throw new Exception(string.Format("{0} does not exist", strReposPath_From));
				}
			}

			_ci.InternalChangeSet_Append(csic);

			if (_args.AutoCommit)
			{
				_ci.InternalChangeSet_SetComment(_args.Comment);
				WriteChangeSet(csic);
				bSuccess = _ci.Commit(csic);
			}
			else
			{
				WriteChangeSet();
			}

			return bSuccess;
		}

		private bool ProcessCommandPin(string strReposPath, int version)
		{
			bool bSuccess = true;

			Login();

			ChangeSetItemColl csic = new ChangeSetItemColl();

			VaultClientFolder vcfolder = null;
			VaultClientFile vcfile = null;

			vcfolder = _ci.TreeCache.Repository.Root.FindFolderRecursive(strReposPath);
			if (vcfolder != null)
			{
				// ok, this is a folder
				VaultObjectVersionInfo[] ovis = null;
				_ci.Connection.GetObjectVersionList(vcfolder.ID, ref ovis, false);
				long objverid = -1;
				foreach (VaultObjectVersionInfo ovi in ovis)
				{
					if (ovi.Version == version)
					{
						objverid = ovi.ObjVerID;
						break;
					}
				}

				if (objverid == -1)
				{
					throw new Exception(string.Format("Version {0} of {1} does not exist", version, vcfolder.FullPath));
				}

				ChangeSetItem_Pin csi = new ChangeSetItem_Pin(
					DateTime.Now,
					_args.Comment,
					String.Empty,
					objverid,
					vcfolder.FullPath);
				csic.Add(csi);
			}
			else
			{
				vcfile = _ci.TreeCache.Repository.Root.FindFileRecursive(strReposPath);
				if (vcfile != null)
				{
					// ok, this is a file
					VaultObjectVersionInfo[] ovis = null;
					_ci.Connection.GetObjectVersionList(vcfile.ID, ref ovis, false);
					long objverid = -1;
					foreach (VaultObjectVersionInfo ovi in ovis)
					{
						if (ovi.Version == version)
						{
							objverid = ovi.ObjVerID;
							break;
						}
					}

					if (objverid == -1)
					{
						throw new Exception(string.Format("Version {0} of {1} does not exist", version, vcfile.FullPath));
					}

					ChangeSetItem_Pin csi = new ChangeSetItem_Pin(
						DateTime.Now,
						_args.Comment,
						String.Empty,
						objverid,
						vcfile.FullPath);
					csic.Add(csi);
				}
				else
				{
					throw new Exception(string.Format("{0} does not exist", strReposPath));
				}
			}

			_ci.InternalChangeSet_Append(csic);

			if (_args.AutoCommit)
			{
				_ci.InternalChangeSet_SetComment(_args.Comment);
				WriteChangeSet(csic);
				bSuccess = _ci.Commit(csic);
			}
			else
			{
				WriteChangeSet();
			}

			return bSuccess;
		}

		private bool ProcessCommandRename(string strReposPath, string strNewName)
		{
			bool bSuccess = true;

			Login();

			ChangeSetItemColl csic = new ChangeSetItemColl();

			VaultClientFolder vcfolder = null;
			VaultClientFile vcfile = null;

			vcfolder = _ci.TreeCache.Repository.Root.FindFolderRecursive(strReposPath);
			if (vcfolder != null)
			{
				// ok, this is a folder
				ChangeSetItem_Rename csi = new ChangeSetItem_Rename(
					DateTime.Now,
					_args.Comment,
					String.Empty,
					vcfolder.ID,
					vcfolder.FullPath,
					strNewName);
				csic.Add(csi);
			}
			else
			{
				vcfile = _ci.TreeCache.Repository.Root.FindFileRecursive(strReposPath);
				if (vcfile != null)
				{
					// ok, this is a file
					ChangeSetItem_Rename csi = new ChangeSetItem_Rename(
						DateTime.Now,
						_args.Comment,
						String.Empty,
						vcfile.ID,
						vcfile.FullPath,
						strNewName);
					csic.Add(csi);
				}
				else
				{
					throw new Exception(string.Format("{0} does not exist", strReposPath));
				}
			}

			_ci.InternalChangeSet_Append(csic);

			if (_args.AutoCommit)
			{
				_ci.InternalChangeSet_SetComment(_args.Comment);
				WriteChangeSet(csic);
				bSuccess = _ci.Commit(csic);
			}
			else
			{
				WriteChangeSet();
			}

			return bSuccess;
		}

		private bool ProcessCommandSetLabelWorkingFolder(string strReposPath, string labelName, string strDiskFolder, string labelSubItem)
		{
			bool bSuccess = true;
			string[] discoveredPaths;
			long labelSubItemId = 0;
			long labelID = 0;
			VaultClientTreeObject labelStructure;
			
			Login();
			_ci.Refresh();

			VaultClientTreeObject reposTreeObj = _ci.TreeCache.Repository.Root.FindTreeObjectRecursive(strReposPath);

			if(reposTreeObj == null)
			{
				throw new Exception(string.Format("{0} does not exist in the repository.", strReposPath));
			}

			labelSubItemId = reposTreeObj.ID;

			// This is the ID of the file/folder where the label structure was
			// created.
			long rootID = 0;

			bSuccess = _ci.GetByLabel_GetStructure(strReposPath, labelName, ref labelID, labelSubItem, out discoveredPaths, out labelStructure, out rootID);

			if(bSuccess == true)
			{
				VaultClientFolder labelRootFolder;

				VaultClientTreeObjectColl treeObjects = new VaultClientTreeObjectColl();

				if (labelStructure is VaultClientFile)
				{
					if (((VaultClientFile) labelStructure).ID == labelSubItemId)
						treeObjects.Add(labelStructure);
				}
				else
				{
					((VaultClientFolder) labelStructure).FindTreeObjectsRecursive(labelSubItemId, ref treeObjects);
				}

				if(treeObjects.Count < 1)
				{
					throw new Exception(string.Format("{0} was not be found in the label", strReposPath));
				}
				else if(treeObjects.Count > 1)
				{
					throw new Exception(string.Format("{0} was not specific within the label", strReposPath));
				}

				if(labelStructure is VaultClientFolder)
				{				
					labelRootFolder = (VaultClientFolder)treeObjects[0];
				}
				else
				{
					// HACK : we need to create a fake parent for this object, since we can't
					// HACK : set a working folder for a file
					labelRootFolder = GetFakeLabelParent((VaultClientFile) labelStructure, strReposPath, labelName);
				}

				if(labelRootFolder == null)
				{
					throw new Exception(string.Format("{0} does not exist in the label structure for {1}", strReposPath, labelName));
				}

				// create the local dir
				if (!Directory.Exists(strDiskFolder))
				{
					Directory.CreateDirectory(strDiskFolder);
				}
				if (!Directory.Exists(strDiskFolder))
				{
					throw new Exception(string.Format("{0} does not exist and could not be created", strDiskFolder));
				}

				// set working folder
				_ci.TreeCache.SetWorkingFolder(labelRootFolder.FullPath, strDiskFolder);

				WriteListWorkingFolders();
			}
			else
			{
				string subItemOptions = String.Empty;
				
				foreach(string item in discoveredPaths)
				{
					subItemOptions += string.Format("   {0}\n", item);
				}

				throw new Exception(string.Format("The specified item is shared to multiple places in the label structure.\n" +
					"Please use \"vault.exe SETLABELWORKINGFOLDER '{0}' '{1}' labelpath\", where labelpath is one of:\n" +
					"\n" +					
					subItemOptions, labelName, strReposPath));
			}

			return bSuccess;
		}

		private bool ProcessCommandShare(string strReposPath_From, string strReposPath_To)
		{
			bool bSuccess = true;
			
			Login();

			ChangeSetItemColl csic = new ChangeSetItemColl();

			VaultClientFolder vcfolder = null;
			VaultClientFile vcfile = null;

			vcfolder = _ci.TreeCache.Repository.Root.FindFolderRecursive(strReposPath_From);
			if (vcfolder != null)
			{
				// ok, this is a folder
				ChangeSetItem_Share csi = new ChangeSetItem_Share(
					DateTime.Now,
					_args.Comment,
					String.Empty,
					vcfolder.ID,
					vcfolder.FullPath,
					strReposPath_To);
				csic.Add(csi);
			}
			else
			{
				vcfile = _ci.TreeCache.Repository.Root.FindFileRecursive(strReposPath_From);
				if (vcfile != null)
				{
					// ok, this is a file
					ChangeSetItem_Share csi = new ChangeSetItem_Share(
						DateTime.Now,
						_args.Comment,
						String.Empty,
						vcfile.ID,
						vcfile.FullPath,
						strReposPath_To);
					csic.Add(csi);
				}
				else
				{
					throw new Exception(string.Format("{0} does not exist", strReposPath_From));
				}
			}

			_ci.InternalChangeSet_Append(csic);

			if (_args.AutoCommit)
			{
				_ci.InternalChangeSet_SetComment(_args.Comment);
				WriteChangeSet(csic);
				bSuccess = _ci.Commit(csic);
			}
			else
			{
				WriteChangeSet();
			}

			return bSuccess;
		}

		private bool ProcessCommandUncloak(ArrayList strItemArray)
		{
			bool bSuccess = true;
			
			Login();
			_ci.Refresh();

			for (int i=0; i < strItemArray.Count; i++)
			{
				string strReposItem = (string) strItemArray[i];

				VaultClientFolder vcfolder = null;

				vcfolder = _ci.TreeCache.Repository.Root.FindFolderRecursive(strReposItem);
				if (vcfolder == null)
				{
					throw new UsageException(string.Format("{0} exists, but the UNCLOAK command can only be used on folders.", strReposItem));
				}
				else
				{
					_ci.CloakOrUnCloak(vcfolder.FullPath, false);
					_xml.WriteUserMessage(string.Format("Cloaked: {0}", vcfolder.FullPath));
				}
			}

			return bSuccess;
		}

		private bool ProcessCommandUndoChangeSetItem(int nChgSetID)
		{
			bool bSuccess = true;

			Login();

			if ( ValidateChangeSetItemID(nChgSetID) == false )
			{
				throw new UsageException(string.Format("Invalid ChangeSetItem ID: {0}.  Please use the LISTCHANGESET command to retrieve a valid ID.", nChgSetID));
			}

			// get the change set item by index - note
			// error checking should have been done at this point.
			ChangeSetItemColl csic = _ci.InternalChangeSet_GetItems(true);
			ChangeSetItem csi = csic[nChgSetID];

			// remove this item from the set.
			_ci.InternalChangeSet_Undo(csi);

			// write the new change set.
			WriteChangeSet();

			return bSuccess;
		}

		private bool ProcessCommandUndoCheckout(string strReposPath)
		{
			bool bSuccess = true;

			Login();

			VaultClientFolder vcfolder = null;
			VaultClientFile vcfile = null;

			vcfolder = _ci.TreeCache.Repository.Root.FindFolderRecursive(strReposPath);
			if (vcfolder != null)
			{
				_ci.UndoCheckOut(vcfolder, _args.Recursive, _args.LocalCopy);
			}
			else
			{
				vcfile = _ci.TreeCache.Repository.Root.FindFileRecursive(strReposPath);

				if (vcfile != null)
				{
					if(_ci.GetWorkingFolder(vcfile.Parent) == null)
					{
						throw new Exception(string.Format("{0} does not have a working folder set", strReposPath));
					}

					_ci.UndoCheckOut(vcfile, _args.LocalCopy);
				}
				else
				{
					throw new Exception(string.Format("{0} does not exist", strReposPath));
				}
			}
			return bSuccess;
		}

		private bool ProcessCommandUnPin(string strReposPath)
		{
			bool bSuccess = true;

			Login();

			ChangeSetItemColl csic = new ChangeSetItemColl();

			VaultClientFolder vcfolder = null;
			VaultClientFile vcfile = null;

			vcfolder = _ci.TreeCache.Repository.Root.FindFolderRecursive(strReposPath);
			if (vcfolder != null)
			{
				// ok, this is a folder
				ChangeSetItem_Unpin csi = new ChangeSetItem_Unpin(
					DateTime.Now,
					_args.Comment,
					String.Empty,
					vcfolder.ID,
					vcfolder.FullPath);
				csic.Add(csi);
			}
			else
			{
				vcfile = _ci.TreeCache.Repository.Root.FindFileRecursive(strReposPath);
				if (vcfile != null)
				{
					// ok, this is a file
					ChangeSetItem_Unpin csi = new ChangeSetItem_Unpin(
						DateTime.Now,
						_args.Comment,
						String.Empty,
						vcfile.ID,
						vcfile.FullPath);
					csic.Add(csi);
				}
				else
				{
					throw new Exception(string.Format("{0} does not exist", strReposPath));
				}
			}

			_ci.InternalChangeSet_Append(csic);

			if (_args.AutoCommit)
			{
				_ci.InternalChangeSet_SetComment(_args.Comment);
				WriteChangeSet(csic);
				bSuccess = _ci.Commit(csic);
			}
			else
			{
				WriteChangeSet();
			}

			return bSuccess;
		}

		private VaultCmdLineClient(Args args, XMLOutputWriter xml)
		{
			_args = args;
			_xml = xml;
		}

		private VaultClientFolder GetFakeLabelParent(VaultClientFile file, string currentPath, string labelName)
		{
			VaultClientFolder fakeParent = new VaultClientFolder();

			System.Security.Cryptography.MD5 md5 = new System.Security.Cryptography.MD5CryptoServiceProvider();
			ASCIIEncoding enc = new ASCIIEncoding();
			
			byte[] currentPathBytes = enc.GetBytes(currentPath.ToLower());
			byte[] labelBytes = enc.GetBytes(labelName.ToLower());

			byte[] combinedBytes = new byte[currentPathBytes.Length + labelBytes.Length + 1];

			byte[] currentPathHash = md5.ComputeHash(currentPathBytes);
			byte[] labelHash = md5.ComputeHash(labelBytes);

			string currentPathMD5 = Convert.ToBase64String(currentPathHash);
			string labelMD5 = Convert.ToBase64String(labelHash);

			fakeParent.Name = string.Format("label:{0}:{1}", currentPathMD5, labelMD5);

			fakeParent.Files.Add(file);
			file.Parent = fakeParent;

			return fakeParent;
		}

		private string GetSessionFileName()
		{
			return Path.Combine(_ci.LocalStoreBasePath, SESSION_FILENAME);
		}
		
		private void StoreSession(string strURLBase, string strUsername, string strPassword, string strRepos)
		{
			try
			{
				TextWriter tw = new StreamWriter(new FileStream(GetSessionFileName(), FileMode.Create, FileAccess.Write, FileShare.None));
				tw.WriteLine(strURLBase);
				tw.WriteLine(strUsername);
				tw.WriteLine(strPassword);
				tw.WriteLine(strRepos);
				tw.Close();
			}
			catch (Exception)
			{
				throw;
			}
		}
		
		private void RetrieveSession(ref string strURLBase, ref string strUsername, ref string strPassword, ref string strRepos)
		{
			try
			{
				TextReader tw = new StreamReader(new FileStream(GetSessionFileName(), FileMode.Open, FileAccess.Read, FileShare.Read));
				strURLBase = tw.ReadLine();
				strUsername = tw.ReadLine();
				strPassword = tw.ReadLine();
				strRepos = tw.ReadLine();
				tw.Close();
			}
			catch (Exception)
			{
			}
		}
		
		private void PurgeSession()
		{
			try
			{
				_ci = new ClientInstance();
				_ci.UseFileSystemWatchers = false;
				_ci.Init(VaultClientNetLib.VaultConnection.AccessLevelType.Client);

				string strFN = GetSessionFileName();
				if ( File.Exists(strFN) == true )
				{
					File.Delete(strFN);
				}
			}
			catch (Exception)
			{
			}
		}

		private void GetLabelTimeStamps(string item, string label1, string label2, out DateTime timestamp1, 
			out DateTime timestamp2)
		{
			timestamp1 = VaultDate.EmptyDate(); // rogue value
			timestamp2 = VaultDate.EmptyDate(); // rogue value

			VaultHistoryQueryRequest hq = new VaultHistoryQueryRequest();
			hq.Recursive = true;

			// search all dates
			hq.BeginDate = VaultDate.EmptyDate();
			hq.EndDate = VaultDate.EmptyDate();
			
			hq.RepID = _ci.ActiveRepositoryID;

			VaultClientFolder vcfolder = null;
			VaultClientFile vcfile = null;

			vcfolder = _ci.TreeCache.Repository.Root.FindFolderRecursive(item);
			if (vcfolder != null)
			{
				hq.TopName = vcfolder.FullPath;
				hq.TopID = vcfolder.ID;
				hq.IsFolder = true;
			}
			else
			{
				vcfile = _ci.TreeCache.Repository.Root.FindFileRecursive(item);
				if (vcfile != null)
				{
					hq.TopName = vcfile.FullPath;
					hq.TopID = vcfile.ID;
					hq.IsFolder = false;
				}
				else
				{
					throw new Exception(string.Format("{0} does not exist", item));
				}
			}

			int nRowsRetrieved = 0;
			string strQryToken = null;

			_ci.Connection.HistoryBegin(hq, _args.HistoryRowLimit, ref nRowsRetrieved, ref strQryToken);
			VaultHistoryItem[] histitems = null;
			_ci.Connection.HistoryFetch(strQryToken, 0, nRowsRetrieved-1, ref histitems);
			_ci.Connection.HistoryEnd(strQryToken);

			foreach (VaultHistoryItem hi in histitems)
			{
				// want labels only
				if ( hi.HistItemType == VaultHistoryType.Label )
				{
					if ( hi.MiscInfo1 == label1 )
					{
						timestamp1 = hi.TxDate;
					}
					else if ( label2.Length > 0 && hi.MiscInfo1 == label2 )
					{
						timestamp2 = hi.TxDate;
					}

					// bail out if we've found both already (or if we've found the first one 
					// and only 1 label requested )
					if ( timestamp1 != VaultDate.EmptyDate() && 
						(timestamp2 != VaultDate.EmptyDate() || label2.Length == 0 ) )
						break;
				}
			}

			// make sure we found all relevant labels
			if ( timestamp1 == VaultDate.EmptyDate() )
				throw new ArgumentException("Invalid label: " + label1);
			else if ( label2.Length > 0 && timestamp2 == VaultDate.EmptyDate() )
				throw new ArgumentException("Invalid label: " + label2);

			return;
		}

		private XmlDocument sortHistory(XmlDocument theHistory)
		{
			XmlDocument doc = new XmlDocument();
			return doc;
		}

		private string DecodeUserRights(uint rights)
		{
			StringBuilder rightsString = new StringBuilder("---");
			if ( (rights & 1) != 0 )
				rightsString[0] = 'R';
			if ( (rights & 2) != 0 )
				rightsString[1] = 'C';
			if ( (rights & 4) != 0 )
				rightsString[2] = 'A';
			return rightsString.ToString();
		}

		private string TranslateActionToString(VaultHistoryItem item)
		{
			byte itemnum = (byte)item.HistItemType;
			string descr = "";
			switch (itemnum)
			{
				#region actions
				case VaultHistoryType.Added:
					descr = "Add";
					break;
				case VaultHistoryType.BranchedFrom:
					descr = "Branch";
					break;
				case VaultHistoryType.BranchedFromShare:
					descr = "BranchedFromShare";
					break;
				case VaultHistoryType.BranchedFromItem:
					descr = "BranchFromItem";
					break;
				case VaultHistoryType.BranchedFromShareItem:
					descr = "BranchFromShareItem";
					break;
				case VaultHistoryType.CheckIn:
					descr = "CheckIn";
					break;
				case VaultHistoryType.Created:
					descr = "Create";
					break;
				case VaultHistoryType.Deleted:
					descr = "Delete";
					break;
				case VaultHistoryType.Label:
					descr = "Label";
					break;
				case VaultHistoryType.MovedFrom:
					descr = "MoveFrom";
					break;
				case VaultHistoryType.MovedTo:
					descr = "MoveTo";
					break;
				case VaultHistoryType.Obliterated:
					descr = "Obliterate";
					break;
				case VaultHistoryType.Pinned:
					descr = "Pin";
					break;
				case VaultHistoryType.PropertyChange:
				#region PropertyChange
					//we get this in the form E:x,M:y where E: marks the change for eol id, and M: marks the change for mergeable. 
					//x will be the new value of the eolid and y will be the new value of the mergeable setting.
					//x or y of 0 means no change and -1 means the value wasn't exp
					try
					{
						string s = item.MiscInfo1;
						string[] props = item.MiscInfo1.Split(',');
						string propName = string.Empty;
						string propValue = string.Empty;
						string[] propNames = null;
						string[] propValues = null;
						descr = "ChangeProperty";

						if (props != null)
						{
							propNames = new string[props.Length];
							propValues = new string[props.Length];
							for (int j=0; j<props.Length; j++)
							{
								string[] tmp = props[j].Split(':');
								if (int.Parse(tmp[1]) == -100)
								{
									propNames[j] = null;
									propValues[j] = null;
								}
								else 
								{
									if (tmp[0] == "E")
									{
										propNames[j] = "EOLConversion";
										switch (int.Parse(tmp[1]))
										{
											case VaultLib.VaultEOLForm.CR:
												propValues[j] = "CR";
												break;
											case VaultLib.VaultEOLForm.CRLF:
												propValues[j] = "CRLF";
												break;
											case VaultLib.VaultEOLForm.LF:
												propValues[j] = "LF";
												break;
											case VaultLib.VaultEOLForm.Native:
												propValues[j] = "Native";
												break;
											case VaultLib.VaultEOLForm.None:
												propValues[j] = "None";
												break;
										}										
									}

									else if (tmp[0] == "M")
									{
										propNames[j] = "FileType";
										switch (int.Parse(tmp[1]))
										{
											case VaultLib.VaultMergeable.Binary:
												propValues[j] = "Binary";
												break;
											case VaultLib.VaultMergeable.Mergeable:
												propValues[j] = "Mergeable";
												break;
											case VaultLib.VaultMergeable.UseSystemDefault:
												propValues[j] = "UseSystemDefault";
												break;														
										}										
									}								
								}
							}						
						}
							
						for (int n = 0; n < propNames.Length; n++)
						{
							if (propNames[n] != null)
							{
								if (descr.Length > 0)
									descr += ", ";

								descr += string.Format("{0} {1} {2}",  propNames[n], "changedTo", propValues[n]);
							}
						}
					}
					catch
					{
						descr = "PropertyChanged";
					}
					
					break;
				#endregion
				case VaultHistoryType.Renamed:
					descr = "RenameItem";
					break;
				case VaultHistoryType.RenamedItem:
					descr = "Rename";
					break;
				case VaultHistoryType.Rollback:
					descr = "Rollback";
					break;
				case VaultHistoryType.SharedTo:
					descr = "Share";
					break;
				case VaultHistoryType.Undeleted:
					descr = "Undelete";
					break;
				case VaultHistoryType.UnPinned:
					descr = "UnPin";
					break;
				case VaultHistoryType.Snapshot:
					descr = "Snapshot";
					break;
				case VaultHistoryType.SnapshotFrom:
					string snapshotName = string.Empty;
					//display only the name of the labeled folder, not the full path
					if (item.MiscInfo2 != null)
					{
						string[] pathItems = item.MiscInfo2.Split('/');
						snapshotName = pathItems[pathItems.Length-1];
					}
					descr = "Snapshot";
					break;
				case VaultHistoryType.SnapshotItem:
					descr = String.Format("{0} {1} {2} ", "Snapshot_", "From", item.MiscInfo1);
					break;
				default:
					descr = "Unknown";
					break;
				#endregion
			}
			return descr;
		}
		private int GetRepositoryId(string repositoryName)
		{
			VaultRepositoryInfo[] reps = null;
			//List all the repositories on the server.
			_ci.ListRepositories(ref reps);

			int repositoryId = -1;

			//Search for the one that we want.
			foreach (VaultRepositoryInfo r in reps)
			{
				if (String.Compare(r.RepName,repositoryName, true) == 0)
				{
					//This will load up the client side cache files and refresh the repository structure.
					//See http://support.sourcegear.com/viewtopic.php?t=6 for more on client side cache files.
					repositoryId = r.RepID;
					break;
				}
			}
			if (repositoryId == -1)
				throw new ArgumentException(string.Format("Repository {0} not found", repositoryName));

			return repositoryId;
		}
		
		#endregion
	}
}
